# -*- coding: utf-8 -*-
"""
Analyzers file for all the different analyzers classes in Deequ
"""
import json

from pyspark.sql import DataFrame, SparkSession, SQLContext

from pydeequ.pandas_utils import ensure_pyspark_df
from pydeequ.repository import MetricsRepository, ResultKey
from enum import Enum
from pydeequ.scala_utils import to_scala_seq
from pydeequ.configs import SPARK_VERSION

class _AnalyzerObject:
    """
    Analyzer base object to pass and accumulate the analyzers of the run with respect to the JVM
    """

    def _set_jvm(self, jvm):
        self._jvm = jvm
        return self

    @property
    def _deequAnalyzers(self):
        if self._jvm:
            return self._jvm.com.amazon.deequ.analyzers
        raise AttributeError(
            "JVM not set, please run _set_jvm() method first."
        )  # TODO: Test that this exception gets raised


class AnalysisRunner:
    """
    Runs a set of analyzers on the data at hand and optimizes the resulting computations to minimize
    the number of scans over the data. Additionally, the internal states of the computation can be
    stored and aggregated with existing states to enable incremental computations.

    :param spark_session SparkSession: SparkSession
    """

    def __init__(self, spark_session: SparkSession):
        self._spark_session = spark_session

    def onData(self, df):
        """
        Starting point to construct an AnalysisRun.
        :param dataFrame df: tabular data on which the checks should be verified
        :return: new AnalysisRunBuilder object
        """
        df = ensure_pyspark_df(self._spark_session, df)
        return AnalysisRunBuilder(self._spark_session, df)


class AnalyzerContext:
    """
    The result returned from AnalysisRunner and Analysis.
    """

    @classmethod
    def successMetricsAsDataFrame(
        cls, spark_session: SparkSession, analyzerContext, forAnalyzers: list = None, pandas: bool = False
    ):
        """
        Get the Analysis Run as a DataFrame.

        :param SparkSession spark_session: SparkSession
        :param AnalyzerContext analyzerContext: Analysis Run
        :param list forAnalyzers: Subset of Analyzers from the Analysis Run
        :return DataFrame: DataFrame of Analysis Run
        """
        if forAnalyzers:
            raise NotImplementedError("forAnalyzers have not been implemented yet.")
        forAnalyzers = getattr(
            spark_session._jvm.com.amazon.deequ.analyzers.runners.AnalyzerContext, "successMetricsAsDataFrame$default$3"
        )()
        analysis_result = (
            spark_session._jvm.com.amazon.deequ.analyzers.runners.AnalyzerContext.successMetricsAsDataFrame(
                spark_session._jsparkSession, analyzerContext, forAnalyzers
            )
        )
        sql_ctx = SQLContext(
            sparkContext=spark_session._sc,
            sparkSession=spark_session,
            jsqlContext=spark_session._jsparkSession.sqlContext(),
        )
        return DataFrame(analysis_result, sql_ctx).toPandas() if pandas else DataFrame(analysis_result, sql_ctx)

    @classmethod
    def successMetricsAsJson(cls, spark_session: SparkSession, analyzerContext, forAnalyzers: list = None):
        """
        Get the Analysis Run as a JSON.

        :param SparkSession spark_session: SparkSession
        :param AnalyzerContext analyzerContext: Analysis Run
        :param list forAnalyzers: Subset of Analyzers from the Analysis Run
        :return JSON : JSON output of Analysis Run
        """
        if forAnalyzers:
            raise NotImplementedError("forAnalyzers have not been implemented yet.")
        forAnalyzers = getattr(
            spark_session._jvm.com.amazon.deequ.analyzers.runners.AnalyzerContext, "successMetricsAsJson$default$2"
        )()
        analysis_result = spark_session._jvm.com.amazon.deequ.analyzers.runners.AnalyzerContext.successMetricsAsJson(
            analyzerContext, forAnalyzers
        )
        return json.loads(analysis_result)


class AnalysisRunBuilder:
    """
    Low level class for running analyzers module. This is meant to be called by AnalysisRunner.

    :param spark_session SparkSession: SparkSession
    :param DataFrame  df: DataFrame to run the Analysis on.
    """

    def __init__(self, spark_session: SparkSession, df: DataFrame):

        self._spark_session = spark_session
        self._jvm = spark_session._jvm
        self._jspark_session = spark_session._jsparkSession
        self._df = df
        self._AnalysisRunBuilder = self._jvm.com.amazon.deequ.analyzers.runners.AnalysisRunBuilder(df._jdf)

    def addAnalyzer(self, analyzer: _AnalyzerObject):
        """
        Adds a single analyzer to the current Analyzer run.

        :param analyzer: Adds an analyzer strategy to the run.
        :return self: for further chained method calls.
        """
        analyzer._set_jvm(self._jvm)
        _analyzer_jvm = analyzer._analyzer_jvm
        self._AnalysisRunBuilder.addAnalyzer(_analyzer_jvm)
        return self

    def run(self):
        """
        Run the Analysis.

        :return: self: Runs the AnalysisRunBuilder.
        """
        return self._AnalysisRunBuilder.run()

    def useRepository(self, repository: MetricsRepository):
        """
        Set a metrics repository associated with the current data to enable features
        like reusing previously computed results and storing the results of the current run.

        :param MetricsRepository repository: A metrics repository to store and
            load results associated with the run
        :return: self
        """
        self._AnalysisRunBuilder = self._AnalysisRunBuilder.useRepository(repository.repository)
        return self

    def saveOrAppendResult(self, resultKey: ResultKey):
        """
        A shortcut to save the results of the run or append them to existing results
        in the metrics repository.

        :param ResultKey resultKey: The result key to identify the current run
        :return: self
        """
        self._AnalysisRunBuilder.saveOrAppendResult(resultKey.resultKey)
        return self


class ApproxCountDistinct(_AnalyzerObject):
    """
    Computes the approximate count distinctness of a column with HyperLogLogPlusPlus.

    :param str column: Column to compute this aggregation on.
    :param str where: Additional filter to apply before the analyzer is run.
    """

    def __init__(self, column: str, where: str = None):
        self.column = column
        self.where = where

    @property
    def _analyzer_jvm(self):
        """Returns the value of the computed aggregation

        :return self
        """
        return self._deequAnalyzers.ApproxCountDistinct(self.column, self._jvm.scala.Option.apply(self.where))


class ApproxQuantile(_AnalyzerObject):
    """
    Computes the Approximate Quantile of a column. The allowed relative error compared
    to the exact quantile can be configured with the `relativeError` parameter.

    :param str column: The column in the DataFrame for which the approximate quantile is analyzed.
    :param float [0,1] quantile: The computed quantile. It must be within the
            interval [0, 1], where 0.5 would be the median.
    :param float [0,1] relativeError: Relative target precision to achieve in the
            quantile computation. A `relativeError` = 0.0 would yield the exact
            quantile while increasing the computational load.
    :param str where: Additional filter to apply before the analyzer is run.
    """

    def __init__(self, column: str, quantile: float, relativeError: float = 0.01, where=None):
        self.column = column
        self.quantile = quantile
        self.relativeError = relativeError
        self.where = where

    @property
    def _analyzer_jvm(self):
        """Returns the value of the computed aggregation

        :return self
        """
        return self._deequAnalyzers.ApproxQuantile(
            self.column, self.quantile, self.relativeError, self._jvm.scala.Option.apply(self.where)
        )


class ApproxQuantiles(_AnalyzerObject):
    """
     Computes the approximate quantiles of a column. The allowed relative
     error compared to the exact quantile can be configured with
     `relativeError` parameter.

    :param str column: Column in DataFrame for which the approximate
            quantile is analyzed.
    :param List[float[0,1]]) quantiles: Computed Quantiles. Must be in
            the interval [0, 1], where 0.5 would be the median.
    :param float [0,1] relativeError: Relative target precision to achieve
            in the quantile computation. A `relativeError` = 0.0 would
            yield the exact quantile while increasing the computational load.
    """

    def __init__(self, column, quantiles, relativeError=0.01):
        self.column = column
        self.quantiles = quantiles
        self.relativeError = relativeError

    @property
    def _analyzer_jvm(self):
        """Returns the value of the computed aggregation

        :return self
        """
        return self._deequAnalyzers.ApproxQuantiles(
            self.column, to_scala_seq(self._jvm, self.quantiles), self.relativeError
        )


class Completeness(_AnalyzerObject):
    """Completeness is the fraction of non-null values in a column.

    :param str column: Column in DataFrame for which Completeness is analyzed.
    :param str where: additional filter to apply before the analyzer is run.
    """

    def __init__(self, column, where=None):
        """
        :param column: Column in DataFrame for which Completeness is analyzed.
        :param str where: additional filter to apply before the analyzer is run.
        """
        self.column = column
        self.where = where

    @property
    def _analyzer_jvm(self):
        """Returns the value of the computed completeness

        :return self: access the value of the Completeness analyzer.
        """
        return self._deequAnalyzers.Completeness(
            self.column,
            self._jvm.scala.Option.apply(self.where),
            self._jvm.scala.Option.apply(None)
        )


class Compliance(_AnalyzerObject):
    """
    Compliance measures the fraction of rows that complies with the given
    column constraint. E.g if the constraint is "att1>3" and data frame
    has 5 rows with att1 column value greater than 3 and 10 rows under
    3; a DoubleMetric would be returned with 0.33 value.

    :param str instance: Unlike other column analyzers (e.g completeness)
        this analyzer can not infer to the metric instance name from
        column name. Also the constraint given here can be referring
        to multiple columns, so metric instance name should be
        provided,describing what the analysis  being done for.
    :param str predicate: SQL-predicate to apply per row
    :param str where: additional filter to apply before
        the analyzer is run.
    """

    def __init__(self, instance, predicate, where=None):

        self.instance = instance
        self.predicate = predicate
        self.where = where

    @property
    def _analyzer_jvm(self):
        """Returns the value of the computed compliance

        :return self
        """
        return self._deequAnalyzers.Compliance(
            self.instance,
            self.predicate,
            self._jvm.scala.Option.apply(self.where),
            self._jvm.scala.collection.Seq.empty(),
            self._jvm.scala.Option.apply(None)
        )


class Correlation(_AnalyzerObject):
    """
    Computes the pearson correlation coefficient between the two given columns.

    :param str column1: First column in the DataFrame for which the Correlation is analyzed.
    :param str column2: Second column in the DataFrame for which the Correlation is analyzed.
    :param str where: additional filter to apply before the analyzer is run.
    """

    def __init__(self, column1, column2, where=None):
        self.column1 = column1
        self.column2 = column2
        self.where = where

    @property
    def _analyzer_jvm(self):
        """Returns the value of the computed correlation

        :return self
        """
        return self._deequAnalyzers.Correlation(self.column1, self.column2, self._jvm.scala.Option.apply(self.where))


class CountDistinct(_AnalyzerObject):
    """
    Counts the distinct elements in the column(s).

    :param List[str] columns: Column(s) in the DataFrame for which distinctness is analyzed.
    """

    def __init__(self, columns):
        if isinstance(columns, str):
            columns = [columns]
        self.columns = columns

    @property
    def _analyzer_jvm(self):
        """Returns the value of the computed distinctness

        :return self
        """
        return self._deequAnalyzers.CountDistinct(to_scala_seq(self._jvm, self.columns))


class DataType(_AnalyzerObject):
    """
    Data Type Analyzer. Returns the datatypes of column

    :param str column: Column in the DataFrame for which data type is analyzed.
    :param str where: additional filter to apply before the analyzer is run.
    """

    def __init__(self, column, where=None):
        self.column = column
        self.where = where

    @property
    def _analyzer_jvm(self):
        """
        Returns the datatype of the column(s)

        :return self
        """
        return self._deequAnalyzers.DataType(self.column, self._jvm.scala.Option.apply(self.where))


class Distinctness(_AnalyzerObject):
    """
    Count the distinctness of elements in column(s).
    Distinctness is the fraction of distinct values of a column(s).

    :param str OR list[str] columns: Column(s) in the DataFrame for which data
        type is to be analyzed. The column is expected to be a str for single
        column or list[str] for multiple columns.
    :param str where: additional filter to apply before the analyzer is run.
    """

    def __init__(self, columns, where: str = None):
        if isinstance(columns, str):
            columns = [columns]
        self.columns = columns
        self.where = where

    @property
    def _analyzer_jvm(self):
        """
        Returns the distinctness of the column(s)

        :return self: access the value of the distincness analyzer.
        """
        return self._deequAnalyzers.Distinctness(
            to_scala_seq(self._jvm, self.columns), self._jvm.scala.Option.apply(self.where)
        )


class Entropy(_AnalyzerObject):
    """
    Entropy is a measure of the level of information contained in a message.
    Given the probability distribution over values in a column, it describes
    how many bits are required to identify a value.

    :param str column: Column in DataFrame for which entropy is calculated.
    :param str where: additional filter to apply before the analyzer is run.
    """

    def __init__(self, column, where: str = None):
        self.column = column
        self.where = where

    @property
    def _analyzer_jvm(self):
        """
        Returns the entropy of the column

        :return self
        """
        return self._deequAnalyzers.Entropy(self.column, self._jvm.scala.Option.apply(self.where))


class Histogram(_AnalyzerObject):
    """
    Histogram is the summary of values in a column of a DataFrame.
    It groups the column's values then calculates the number of rows with
    that specific value and the fraction of the value.

    :param str column: Column in DataFrame to do histogram analysis.
    :param lambda expr binningUdf: Optional binning function to run before
        grouping to re-categorize the column values.For example to turn a
        numerical value to a categorical value binning functions might be used.
    :param int maxDetailBins: Histogram details is only provided for N column
        values with top counts. MaxBins sets the N. This limit does not affect
        what is being returned as number of bins.It always returns the distinct
        value count.
    :param str where: additional filter to apply before the analyzer is run.
    """

    def __init__(self, column, binningUdf=None, maxDetailBins: int = None, where: str = None):

        self.column = column
        self.binningUdf = binningUdf
        self.maxDetailBins = maxDetailBins
        self.where = where

    @property
    def _analyzer_jvm(self):
        """Returns the histogram summary of values in a column.

        :return self
        """
        if not self.maxDetailBins:
            self.maxDetailBins = getattr(self._jvm.com.amazon.deequ.analyzers.Histogram, "apply$default$3")()
        return self._deequAnalyzers.Histogram(
            self.column,
            self._jvm.scala.Option.apply(self.binningUdf),
            self.maxDetailBins,
            self._jvm.scala.Option.apply(self.where),
            getattr(self._jvm.com.amazon.deequ.analyzers.Histogram, "apply$default$5")(),
            getattr(self._jvm.com.amazon.deequ.analyzers.Histogram, "apply$default$6")()
        )


class KLLParameters:
    """
    Parameter definition for KLL Sketches.

    :param int sketchSize: size of kll sketch.
    :param float shrinkingFactor: shrinking factor of kll sketch.
    :param int numberOfBuckets:  number of buckets.
    """

    def __init__(self, spark_session: SparkSession, sketchSize: int, shrinkingFactor: float, numberOfBuckets: int):
        self._spark_session = spark_session
        self.sketchSize = sketchSize
        self.shrinkingFactor = shrinkingFactor
        self.numberOfBuckets = numberOfBuckets

    @property
    def _param(self):
        """
        Return the JVM KLLParameter object
        """
        return self._spark_session._jvm.com.amazon.deequ.analyzers.KLLParameters(
            self.sketchSize, self.shrinkingFactor, self.numberOfBuckets
        )


class KLLSketch(_AnalyzerObject):
    """
    The KLL Sketch analyzer.

    :param str column: Column in DataFrame to do histogram analysis.
    :param KLLParameters kllParameters: parameters of KLL Sketch
    """

    def __init__(self, column: str, kllParameters: KLLParameters):
        self.column = column
        self.kllParameters = kllParameters

    @property
    def _analyzer_jvm(self):
        """
        Returns the histogram summary of values in a column.

        :return self
        """
        if not self.kllParameters:
            self.kllParameters = getattr(self._jvm.com.amazon.deequ.analyzers.KLLSketch, "apply$default$2")()
        return self._deequAnalyzers.KLLSketch(self.column, self._jvm.scala.Option.apply(self.kllParameters._param))


class Maximum(_AnalyzerObject):
    """Get the maximum of a numeric column."""

    def __init__(self, column, where: str = None):
        """
        :param str column: column to find the maximum.
        :param str where: additional filter to apply before the analyzer is run.
        """
        self.column = column
        self.where = where

    @property
    def _analyzer_jvm(self):
        """Returns the maximum value in a column.

        :return self
        """
        return self._deequAnalyzers.Maximum(
            self.column, self._jvm.scala.Option.apply(self.where), self._jvm.scala.Option.apply(None)
        )


class MaxLength(_AnalyzerObject):
    """MaxLength Analyzer. Get Max length of a str type column.

    :param str column: column in DataFrame to find the maximum length.
            Column is expected to be a str type.
    :param str where: additional filter to apply before the analyzer is run.
    """

    def __init__(self, column, where: str = None):

        self.column = column
        self.where = where

    @property
    def _analyzer_jvm(self):
        """Returns the maximum length in a column.

        :return self
        """
        return self._deequAnalyzers.MaxLength(
            self.column,
            self._jvm.scala.Option.apply(self.where),
            self._jvm.scala.Option.apply(None)
        )


class Mean(_AnalyzerObject):
    """
    Mean Analyzer. Get mean of a column

    :param str column: Column in DataFrame to find the mean.
    :param str where: additional filter to apply before the analyzer is run.
    """

    def __init__(self, column, where: str = None):

        self.column = column
        self.where = where

    @property
    def _analyzer_jvm(self):
        """
        Returns the mean of a column.

        :return self
        """
        return self._deequAnalyzers.Mean(self.column, self._jvm.scala.Option.apply(self.where))


class Minimum(_AnalyzerObject):
    """Count the distinct elements in a single or multiple columns

    :param str column: Column in DataFrame to find the minimum value.
    :param str where: additional filter to apply before the analyzer is run.
    """

    def __init__(self, column, where: str = None):
        self.column = column
        self.where = where

    @property
    def _analyzer_jvm(self):
        """Returns the minimum of a column.

        :return self
        """
        return self._deequAnalyzers.Minimum(
            self.column, self._jvm.scala.Option.apply(self.where), self._jvm.scala.Option.apply(None)
        )


class MinLength(_AnalyzerObject):
    """
    Get the minimum length of a column

    :param str column: Column in DataFrame to find the minimum Length.
        Column is expected to be a str type.
    :param str where : additional filter to apply before the analyzer is run.
    """

    def __init__(self, column, where: str = None):
        self.column = column
        self.where = where

    @property
    def _analyzer_jvm(self):
        """
        Returns the minimum length of column.

        :return self
        """
        return self._deequAnalyzers.MinLength(
            self.column,
            self._jvm.scala.Option.apply(self.where),
            self._jvm.scala.Option.apply(None)
        )


class MutualInformation(_AnalyzerObject):
    """
    Describes how much information about one column can be inferred from another column.

    :param list[str] columns: Columns in DataFrame for mutual information analysis.
    :param str where: additional filter to apply before the analyzer is run.
    """

    def __init__(self, columns, where: str = None):
        self.columns = columns
        self.where = where

    @property
    def _analyzer_jvm(self):
        """
        Returns the mutual information of columns.

        :return self
        """
        return self._deequAnalyzers.MutualInformation(
            to_scala_seq(self._jvm, self.columns), self._jvm.scala.Option.apply(self.where)
        )


class PatternMatch(_AnalyzerObject):
    """
    PatternMatch is a measure of the fraction of rows that complies with a
    given column regex constraint.

    E.g A sample dataFrame column has five rows that contain a credit card
    number and 10 rows that do not. According to regex, using the
    constraint Patterns.CREDITCARD returns a doubleMetric .33 value.


    :param str column: Column in DataFrame to check pattern.
    :param str pattern_regex: pattern regex
    :param str pattern_groupNames: groupNames for pattern regex
    :param str where: additional filter to apply before the analyzer is run.
    """

    def __init__(self, column, pattern_regex: str, *pattern_groupNames, where: str = None):
        self.column = column
        self.pattern_regex = pattern_regex
        if pattern_groupNames:
            raise NotImplementedError("pattern_groupNames have not been implemented yet.")
        self.pattern_groupNames = None
        self.where = where

    @property
    def _analyzer_jvm(self):
        """
        Returns the pattern match of column.

        :return self
        """
        return self._deequAnalyzers.PatternMatch(
            self.column,
            self._jvm.scala.util.matching.Regex(self.pattern_regex, None),
            # TODO: revisit bc scala constructor does some weird implicit type casting from python str -> java list
            #  if we don't cast it to str()
            self._jvm.scala.Option.apply(self.where),
            self._jvm.scala.Option.apply(None)
        )


class Size(_AnalyzerObject):
    """
    Size is the number of rows in a DataFrame.

    :param str where: additional filter to apply before the analyzer is run.
    """

    def __init__(self, where: str = None):
        self.where = where

    @property
    def _analyzer_jvm(self):
        """
        Returns the size of DataFrame.

        :return self
        """
        return self._deequAnalyzers.Size(self._jvm.scala.Option.apply(self.where))


class StandardDeviation(_AnalyzerObject):
    """
    Calculates the Standard Deviation of column

    :param str column: Column in DataFrame for standard deviation calculation.
    :param str where: additional filter to apply before the analyzer is run.
    """

    def __init__(self, column, where: str = None):
        self.column = column
        self.where = where

    @property
    def _analyzer_jvm(self):
        """
        Returns the standard deviation of column.

        :return self
        """
        return self._deequAnalyzers.StandardDeviation(self.column, self._jvm.scala.Option.apply(self.where))


class Sum(_AnalyzerObject):
    """
    Calculates the sum of a column

    :param str column: Column in DataFrame to calculate the sum.
    :param str where: additional filter to apply before the analyzer is run.
    """

    def __init__(self, column, where: str = None):

        self.column = column
        self.where = where

    @property
    def _analyzer_jvm(self):
        """
        Returns the sum of column.

        :return self
        """
        return self._deequAnalyzers.Sum(self.column, self._jvm.scala.Option.apply(self.where))


class Uniqueness(_AnalyzerObject):
    """
    Uniqueness is the fraction of unique values of column(s),
    values that occur exactly once.

    :param list[str] columns: Columns in DataFrame to find uniqueness.
    :param str where: additional filter to apply before the analyzer is run.
    """

    def __init__(self, columns, where: str = None):
        self.columns = columns
        self.where = where

    @property
    def _analyzer_jvm(self):
        """
        Returns the uniqueness in columns.

        :return self
        """
        return self._deequAnalyzers.Uniqueness(
            to_scala_seq(self._jvm, self.columns),
            self._jvm.scala.Option.apply(self.where),
            self._jvm.scala.Option.apply(None)
        )


class UniqueValueRatio(_AnalyzerObject):
    """
    Calculates the ratio of uniqueness.

    :param list[str] columns: Columns in DataFrame to find unique value ratio.
    :param str where: additional filter to apply before the analyzer is run.

    """

    def __init__(self, columns, where: str = None):
        self.columns = columns
        self.where = where

    @property
    def _analyzer_jvm(self):
        """
        Returns the unique value ratio in columns.

        :return self
        """
        return self._deequAnalyzers.UniqueValueRatio(
            to_scala_seq(self._jvm, self.columns),
            self._jvm.scala.Option.apply(self.where),
            self._jvm.scala.Option.apply(None)
        )

class DataTypeInstances(Enum):
    """
    An enum class that types columns to scala datatypes
    """
    Boolean = "Boolean"
    Unknown = "Unknown"
    Fractional = "Fractional"
    Integral = "Integral"
    String = "String"

    def _create_java_object(self, jvm):
        dataType_analyzers_class = jvm.com.amazon.deequ.analyzers.DataTypeInstances
        if self == DataTypeInstances.String:
            return dataType_analyzers_class.String()
        elif self == DataTypeInstances.Boolean:
            return dataType_analyzers_class.Boolean()
        elif self == DataTypeInstances.Unknown:
            return dataType_analyzers_class.Unknown()
        elif self == DataTypeInstances.Integral:
            return dataType_analyzers_class.Integral()
        elif self == DataTypeInstances.Fractional:
            return dataType_analyzers_class.Fractional()
        else:
            raise ValueError(f"{jvm} is not a valid datatype Object")


from pydeequ.analyzers import _AnalyzerObject
import json

from pyspark import SQLContext
from pyspark.sql import DataFrame, SparkSession

from pydeequ.analyzers import AnalysisRunBuilder
from pydeequ.checks import Check
from pydeequ.pandas_utils import ensure_pyspark_df

# TODO integrate Analyzer context


class AnomalyCheckConfig:
    pass


class VerificationResult:
    """The results returned from the VerificationSuite
    :param verificationRunBuilder verificationRun:  verification result run()
    """

    def __init__(self, spark_session: SparkSession, verificationRun):
        self._spark_session = spark_session
        self.verificationRun = verificationRun

    @property
    def status(self):
        """
        Getter for the overall status of the verification Run()
        :return: gets the overall status of the verification Run()
        """
        return str(self.verificationRun.status())

    @property
    def checkResults(self):
        """
        Getter for the checks and checksResults of the verification Run()
        :return: JSON with the checks and the checkResults of the verification Run()
        """
        return self.checkResultsAsJson(self._spark_session, self)

    @property
    def metrics(self):
        """
        Getter for the analyzers and their metric results of the verification Run()
        :return: gets the Analyzers and their metric results of the verification Run()
        """
        return self.verificationRun.metrics()

    @classmethod
    def successMetricsAsDataFrame(
        cls, spark_session: SparkSession, verificationResult, forAnalyzers: list = None, pandas: bool = False
    ):
        """
        The results returned in a Data Frame

        :param spark_session: Sparksession
        :param verificationResult: Result of the verification Run()
        :param forAnalyzers: Subset of Analyzers from the Analysis Run
        :return: Data frame of the verification Run()
        """
        if forAnalyzers:
            raise NotImplementedError("forAnalyzers have not been implemented yet.")
        forAnalyzers = getattr(
            spark_session._jvm.com.amazon.deequ.VerificationResult, "successMetricsAsDataFrame$default$3"
        )()

        df = spark_session._jvm.com.amazon.deequ.VerificationResult.successMetricsAsDataFrame(
            spark_session._jsparkSession, verificationResult.verificationRun, forAnalyzers
        )

        sql_ctx = SQLContext(
            sparkContext=spark_session._sc,
            sparkSession=spark_session,
            jsqlContext=spark_session._jsparkSession.sqlContext(),
        )
        return DataFrame(df, sql_ctx).toPandas() if pandas else DataFrame(df, sql_ctx)

    @classmethod
    def successMetricsAsJson(cls, spark_session: SparkSession, verificationResult, forAnalyzers: list = None):
        """
        The results returned in a JSON

        :param SparkSession spark_session: Sparksession
        :param verificationResult: Result of the verification Run()
        :param forAnalyzers: Subset of Analyzers from the Analysis Run
        :return: JSON of the verification Run()
        """
        if forAnalyzers:
            raise NotImplementedError("forAnalyzers have not been implemented yet.")
        forAnalyzers = getattr(
            spark_session._jvm.com.amazon.deequ.VerificationResult, "successMetricsAsJson$default$2"
        )()
        df = spark_session._jvm.com.amazon.deequ.VerificationResult.successMetricsAsJson(
            verificationResult.verificationRun, forAnalyzers
        )
        return json.loads(df)

    @classmethod
    def checkResultsAsJson(cls, spark_session: SparkSession, verificationResult, forChecks=None):
        """
         Returns the check verification Results as a JSON

        :param SparkSession spark_session: SparkSession
        :param verificationResult: The results of the verification run
        :param forChecks: Subset of Checks
        :return: returns a JSON
        """
        if forChecks:
            raise NotImplementedError("forChecks have not been implemented yet.")
        forChecks = getattr(spark_session._jvm.com.amazon.deequ.VerificationResult, "checkResultsAsJson$default$2")()
        df = spark_session._jvm.com.amazon.deequ.VerificationResult.checkResultsAsJson(
            verificationResult.verificationRun, forChecks
        )
        return json.loads(df)

    @classmethod
    def checkResultsAsDataFrame(
        cls, spark_session: SparkSession, verificationResult, forChecks=None, pandas: bool = False
    ):
        """
        Returns the verificaton Results as a Data Frame

        :param SparkSession spark_session: SparkSession
        :param verificationResult: The results of the verification run
        :param forChecks: Subset of Checks
        :return: returns a Data Frame of the results
        """
        if forChecks:
            raise NotImplementedError("forChecks have not been implemented yet.")
        forChecks = getattr(
            spark_session._jvm.com.amazon.deequ.VerificationResult, "checkResultsAsDataFrame$default$3"
        )()

        df = spark_session._jvm.com.amazon.deequ.VerificationResult.checkResultsAsDataFrame(
            spark_session._jsparkSession, verificationResult.verificationRun, forChecks
        )
        sql_ctx = SQLContext(
            sparkContext=spark_session._sc,
            sparkSession=spark_session,
            jsqlContext=spark_session._jsparkSession.sqlContext(),
        )
        return DataFrame(df, sql_ctx).toPandas() if pandas else DataFrame(df, sql_ctx)


class VerificationRunBuilder:
    # TODO Remaining Methods

    """A class to build a Verification Run"""

    def __init__(self, spark_session: SparkSession, data: DataFrame):
        """
        :param SparkSession spark_session: SparkSession
        :param DataFrame data: DataFrame of the data
        """
        if not isinstance(spark_session, SparkSession):
            raise TypeError(f"Expected SparkSession object for spark_session, not {type(spark_session)}")
        if not isinstance(data, DataFrame):
            raise TypeError(f"Expected DataFrame object for data, not {type(data)}")

        self._spark_session = spark_session
        self._sc = spark_session.sparkContext
        self._jvm = spark_session._jvm
        self._VerificationRunBuilder = self._jvm.com.amazon.deequ.VerificationRunBuilder(data._jdf)
        self.checks = []

    def addCheck(self, check: Check):
        """
        Adds checks to the run using checks method.

        :param Check check: A check object to be executed during the run
        :return: Adds checks to the run
        """
        # TODO Support Multiple checks
        self.checks.append(check)
        self._VerificationRunBuilder.addCheck(check._Check)
        return self

    def addAnomalyCheck(self, anomaly, analyzer: _AnalyzerObject, anomalyCheckConfig=None):
        """
        Add a check using anomaly_detection methods. The Anomaly Detection Strategy only checks
        if the new value is an Anomaly.

        :param anomaly:The anomaly detection strategy
        :param AnalysisRunBuilder analyzer: The analyzer for the metric to run anomaly detection on
        :param anomalyCheckConfig: Some configuration settings for the Check
        :return: Adds an anomaly strategy to the run
        """
        if anomalyCheckConfig:
            raise NotImplementedError("anomalyCheckConfigs have not been implemented yet, using default value")

        AnomalyCheckConfig = self._jvm.scala.Option.apply(anomalyCheckConfig)

        anomaly._set_jvm(self._jvm)
        anomaly_jvm = anomaly._anomaly_jvm

        analyzer._set_jvm(self._jvm)
        analyzer_jvm = analyzer._analyzer_jvm

        self._VerificationRunBuilder.addAnomalyCheck(anomaly_jvm, analyzer_jvm, AnomalyCheckConfig)
        return self

    def run(self):
        """
        A method that runs the desired VerificationRunBuilder functions on the data to obtain a Verification Result
        :return:a verificationResult object
        """
        return VerificationResult(self._spark_session, self._VerificationRunBuilder.run())

    def useRepository(self, repository):
        """
        This method reassigns our AnalysisRunBuilder because useRepository returns back a different
        class: AnalysisRunBuilderWithRepository

        Sets a metrics repository associated with the current data to enable features like reusing previously computed
        results and storing the results of the current run.

        :param repository: a metrics repository to store and load results associated with the run
        """
        self._VerificationRunBuilder = self._VerificationRunBuilder.useRepository(repository.repository)
        return self

    def saveOrAppendResult(self, resultKey):
        """
        A shortcut to save the results of the run or append them to the existing results in the metrics repository.

        :param resultKey: The result key to identify the current run
        :return: :A VerificationRunBuilder.scala object that saves or appends a result
        """
        self._VerificationRunBuilder.saveOrAppendResult(resultKey.resultKey)
        return self


class VerificationRunBuilderWithSparkSession(VerificationRunBuilder):
    pass


class VerificationSuite:
    # TODO Remaining Methods
    """Responsible for running checks, the required analysis and return the results"""

    def __init__(self, spark_session):
        """

        :param SparkSession spark_session: The SparkSession
        """
        if not isinstance(spark_session, SparkSession):
            raise TypeError(f"Expected SparkSession object for spark_session, not {type(spark_session)}")

        self._spark_session = spark_session
        self._sc = spark_session.sparkContext
        self._jvm = spark_session._jvm

    def onData(self, df):
        """
        Starting point to construct a VerificationRun.
        :param data:  Tabular data on which the checks should be verified
        :return: The starting point to construct a verificationRun
        """
        df = ensure_pyspark_df(self._spark_session, df)
        return VerificationRunBuilder(self._spark_session, df)



# -*- coding: utf-8 -*-
"""
Suggestions file for all the Constraint Suggestion classes in Deequ

Author: Calvin Wang
"""
import json

from pyspark.sql import DataFrame, SparkSession

from pydeequ.pandas_utils import ensure_pyspark_df
from pydeequ.configs import IS_DEEQU_V1
from pydeequ.scala_utils import scala_get_default_argument


class ConstraintSuggestionRunner:
    """
    Primary class for interacting with suggestions module

    :param SparkSession spark_session: sparkSession
    """

    def __init__(self, spark_session: SparkSession):
        self._spark_session = spark_session

    def onData(self, df):
        """
        Starting point to construct suggestions

        :param df: Tabular data on which the suggestions module use
        :return: The starting point to construct a suggestions module
        """
        df = ensure_pyspark_df(self._spark_session, df)
        return ConstraintSuggestionRunBuilder(self._spark_session, df)


class ConstraintSuggestionRunBuilder:
    """
    Low level class for running suggestions module

    :param SparkSession spark_session: sparkSession
    :param df: Tabular data on which the suggestions module use
    """

    def __init__(self, spark_session: SparkSession, df: DataFrame):
        self._spark_session = spark_session
        self._jvm = spark_session._jvm
        self._jspark_session = spark_session._jsparkSession
        self._df = df
        self._ConstraintSuggestionRunBuilder = (
            self._jvm.com.amazon.deequ.suggestions.ConstraintSuggestionRunBuilder(df._jdf)
        )

    def addConstraintRule(self, constraintRule):
        """
        Add a single rule for suggesting constraints based on ColumnProfiles to the run.

        :param ConstraintRule constraintRule:  A rule that the dataset will be evaluated on throughout to the run.
                To run all the rules on the dataset use .addConstraintRule(DEFAULT())

        :return  self for further method calls.
        """
        constraintRule._set_jvm(self._jvm)
        constraintRule_jvm = constraintRule.rule_jvm
        if isinstance(constraintRule_jvm, list):
            for rule in constraintRule_jvm:
                rule._set_jvm(self._jvm)
                rule_jvm = rule.rule_jvm
                self._ConstraintSuggestionRunBuilder.addConstraintRule(rule_jvm)

        else:
            self._ConstraintSuggestionRunBuilder.addConstraintRule(constraintRule_jvm)

        return self

    def run(self):
        """
        A method that runs the desired ConstraintSuggestionRunBuilder functions on the data to obtain a constraint
                suggestion result. The result is then translated to python.

        :return: A constraint suggestion result
        """
        result = self._ConstraintSuggestionRunBuilder.run()

        jvmSuggestionResult = self._jvm.com.amazon.deequ.suggestions.ConstraintSuggestionResult
        result_json = json.loads(jvmSuggestionResult.getConstraintSuggestionsAsJson(result))
        for cs in result_json["constraint_suggestions"]:
            cs["code_for_constraint"] = self.__s2p_filter(cs["code_for_constraint"])
        return result_json

    @staticmethod
    def __s2p_filter(code: str):
        """
        Scala -> Python translator for the constraint suggestions code
        A method that returns the python translation of the scala constraint suggestion

        :param str code: a scala constraint suggestion

        :return code that is translated to look more like python code
        """
        if " _ " in code:
            code = code.replace(" _ ", " lambda x: x ")

        if "Some(" in code:
            # Usually at the end as 'where' or 'hint' strings as optional
            code = code.replace("Some(", "")[:-1]

        if "Array(" in code:
            # TODO: what if multiple?
            # TODO: Probz redo with regex
            start = code.index("Array(") + len("Array(")
            for idx in range(start, len(code)):
                if code[idx] == ")":
                    code = code[:idx] + "]" + code[idx + 1 :]
                    code = code.replace("Array(", "[")
                    break

        if "Seq(" in code:
            # TODO: what if multiple?
            # TODO: Probz redo with regex
            start = code.index("Seq(") + len("Seq(")
            for idx in range(start, len(code)):
                if code[idx] == ")":
                    code = code[:idx] + "]" + code[idx + 1 :]
                    code = code.replace("Seq(", "[")
                    break

        return code


class ConstraintSuggestion:

    pass


class ConstraintSuggestionResult:
    """
    The result returned from the ConstraintSuggestionSuite
    """


class _RulesObject:
    """Rules base object to pass and accumulate the rules of the run with respect to the JVM"""

    def _set_jvm(self, jvm):

        self._jvm = jvm
        return self

    @property
    def _deequSuggestions(self):
        if self._jvm:
            return self._jvm.com.amazon.deequ.suggestions
        raise AttributeError(
            "JVM not set, please run _set_jvm() method first."
        )  # TODO: Test that this exception gets raised


class DEFAULT(_RulesObject):
    """
    DEFAULT runs all the rules on the dataset.
    """

    @property
    def rule_jvm(self):
        """

        :return:
        """
        return [
            CategoricalRangeRule(),
            CompleteIfCompleteRule(),
            FractionalCategoricalRangeRule(),
            NonNegativeNumbersRule(),
            RetainCompletenessRule(),
            RetainTypeRule(),
            UniqueIfApproximatelyUniqueRule(),
        ]


class CategoricalRangeRule(_RulesObject):
    """
    If we see a categorical range for a column, we suggest an IS IN (...) constraint
    """

    @property
    def rule_jvm(self):
        if IS_DEEQU_V1:
            return self._deequSuggestions.rules.CategoricalRangeRule()

        # DISCLAIMER: this is a workaround for using the default category sorter
        default_category_sorter = scala_get_default_argument(
            self._deequSuggestions.rules.CategoricalRangeRule, 1
        )
        return self._deequSuggestions.rules.CategoricalRangeRule(default_category_sorter)


class CompleteIfCompleteRule(_RulesObject):
    """
    If a column is complete in the sample, we suggest a NOT NULL constraint
    """

    @property
    def rule_jvm(self):
        return self._deequSuggestions.rules.CompleteIfCompleteRule()


class FractionalCategoricalRangeRule(_RulesObject):
    """
    If we see a categorical range for most values in a column, we suggest an IS IN (...)
    constraint that should hold for most values

    :param float targetDataCoverageFraction: The numerical fraction for which the data should stay within range
    of eachother
    """

    def __init__(self, targetDataCoverageFraction: float = 0.9):

        self.targetDataCoverageFraction = targetDataCoverageFraction

    @property
    def rule_jvm(self):
        if IS_DEEQU_V1:
            return self._deequSuggestions.rules.FractionalCategoricalRangeRule(
                self.targetDataCoverageFraction
            )

        # DISCLAIMER: this is a workaround for using the default category sorter
        default_category_sorter = scala_get_default_argument(
            self._deequSuggestions.rules.FractionalCategoricalRangeRule, 2
        )
        return self._deequSuggestions.rules.FractionalCategoricalRangeRule(
            self.targetDataCoverageFraction, default_category_sorter
        )


class NonNegativeNumbersRule(_RulesObject):
    """
    If we see only non-negative numbers in a column, we suggest a corresponding constraint
    """

    @property
    def rule_jvm(self):
        return self._deequSuggestions.rules.NonNegativeNumbersRule()


class RetainCompletenessRule(_RulesObject):
    """
    If a column is incomplete in the sample, we model its completeness as a binomial variable,
    estimate a confidence interval and use this to define a lower bound for the completeness
    """

    @property
    def rule_jvm(self):
        return self._deequSuggestions.rules.RetainCompletenessRule()


class RetainTypeRule(_RulesObject):
    """
    If we detect a non-string type, we suggest a type constraint
    """

    @property
    def rule_jvm(self):
        return self._deequSuggestions.rules.RetainTypeRule()


class UniqueIfApproximatelyUniqueRule(_RulesObject):
    """
    If the ratio of approximate num distinct values in a column is close to the number of records
    (within error of HLL sketch), we suggest a UNIQUE constraint
    """

    @property
    def rule_jvm(self):
        return self._deequSuggestions.rules.UniqueIfApproximatelyUniqueRule()



# -*- coding: utf-8 -*-
from enum import Enum

######################## BASE ANALYZER CLASS #######################


class _AnomalyObject:
    """AnomalyDetection object init to pass and accumulate the anomaly strategies of the run"""

    def _set_jvm(self, jvm):
        self._jvm = jvm
        return self

    @property
    def _deequAnomalies(self):
        if self._jvm:
            return self._jvm.com.amazon.deequ.anomalydetection
        raise AttributeError(
            "JVM not set, please run _set_jvm() method first."
        )  # TODO: Test that this exception gets raised


######################## INDIVIDUAL ANOMALY STRATEGIES  ########################


class RelativeRateOfChangeStrategy(_AnomalyObject):
    """
    Detects anomalies based on the values' rate of change.
    For example RelativeRateOfChangeStrategy(.9, 1.1, 1) calculates the first discrete difference,
    and if some point's value changes by more than 10.0 percent in one timestep it flags it as an anomaly.

    :param float maxRateDecrease: Lower bound of accepted relative change (as new value / old value).
    :param float maxRateIncrease: Upper bound of accepted relative change (as new value / old value).
    :param int order: Order of the calculated difference.
            Defaulted to 1, it calculates the difference between two consecutive
            values. The order of the difference can be set manually. If it is set to 0, this strategy acts
            like the SimpleThresholdStrategy class.
    """

    def __init__(self, maxRateDecrease=None, maxRateIncrease=None, order=1):

        self.maxRateDecrease = maxRateDecrease
        self.maxRateIncrease = maxRateIncrease
        self.order = order

    @property
    def _anomaly_jvm(self):
        return self._deequAnomalies.RelativeRateOfChangeStrategy(
            self._jvm.scala.Option.apply(self.maxRateDecrease),
            self._jvm.scala.Option.apply(self.maxRateIncrease),
            self.order,
        )


class AbsoluteChangeStrategy(_AnomalyObject):
    """
    Detects anomalies based on the values' absolute change.
    AbsoluteChangeStrategy(-10.0, 10.0, 1) for example calculates the first discrete difference
    and if some point's value changes by more than 10.0 in one timestep, it flags it as an anomaly.

    :param float maxRateDecrease: Upper bound of accepted decrease (lower bound of increase).
    :param float maxRateIncrease: Upper bound of accepted growth.
    :param int order:  Order of the calculated difference.
            Defaulted to 1, it calculates the difference between two consecutive
            values. The order of the difference can be set manually. If it is set to 0, this strategy acts
            like the SimpleThresholdStrategy class.
    """

    def __init__(self, maxRateDecrease=None, maxRateIncrease=None, order=1):

        self.maxRateDecrease = maxRateDecrease
        self.maxRateIncrease = maxRateIncrease
        self.order = order

    @property
    def _anomaly_jvm(self):
        return self._deequAnomalies.AbsoluteChangeStrategy(
            self._jvm.scala.Option.apply(self.maxRateDecrease),
            self._jvm.scala.Option.apply(self.maxRateIncrease),
            self.order,
        )


class SimpleThresholdStrategy(_AnomalyObject):
    """A simple anomaly detection strategy that checks if values are  in a specified range.

    :param float lowerBound: Lower bound of accepted range of values
    :param float upperBound: Upper bound of accepted range of values
    """

    def __init__(self, lowerBound, upperBound):

        self.lowerBound = float(lowerBound)
        self.upperBound = float(upperBound)

    @property
    def _anomaly_jvm(self):
        return self._deequAnomalies.SimpleThresholdStrategy(self.lowerBound, self.upperBound)


class RateOfChangeStrategy(_AnomalyObject):
    """
    @Deprecated
    The old RateOfChangeStrategy actually detects absolute changes so it has been migrated to use the
    AbsoluteChangeStrategy class. Use RelativeRateOfChangeStrategy if you want to detect relative changes to the
    previous values.

    """

    def __init__(self, maxRateDecrease=None, maxRateIncrease=None, order=1):
        # self.maxRateDecrease = maxRateDecrease
        # self.maxRateIncrease = maxRateIncrease
        # self.order = order
        pass

    @property
    def _anomaly_jvm(self):
        raise DeprecationWarning("use AbsoluteChangeStrategy instead which describes the strategy more accurately")


class OnlineNormalStrategy(_AnomalyObject):
    """
    Detects anomalies based on the running mean and standard deviation. Anomalies can be excluded from the computation to
    not affect the calculated mean/ standard deviation. This strategy assumes that the data is normally distributed.

    :param float lowerDeviationFactor: Detects anomalies if they are smaller than the mean - lowerDeviationFactor *stdDev
    :param float upperDeviationFactor: Detect anomalies if they are bigger than mean + upperDeviationFactor * stDev
    :param float ignoreStartPercentage: Percentage of data points after start in which no anomalies should be detected
                        (mean and stdDev are probably not representative before).
    :param boolean ignoreAnomalies: If set to true, ignores anomalous points in mean and variance calculation
    """

    def __init__(
        self,
        lowerDeviationFactor=3.0,
        upperDeviationFactor=3.0,
        ignoreStartPercentage=0.1,
        ignoreAnomalies=True,
    ):
        self.lowerDeviationFactor = lowerDeviationFactor
        self.upperDeviationFactor = upperDeviationFactor
        self.ignoreStartPercentage = ignoreStartPercentage
        self.ignoreAnomalies = ignoreAnomalies

    # TODO: Consider adding OnlineCalculationResult
    @property
    def _anomaly_jvm(self):
        return self._deequAnomalies.OnlineNormalStrategy(
            self._jvm.scala.Option.apply(self.lowerDeviationFactor),
            self._jvm.scala.Option.apply(self.upperDeviationFactor),
            self.ignoreStartPercentage,
            self.ignoreAnomalies,
        )


class BatchNormalStrategy(_AnomalyObject):
    """
    Detects anomalies based on the mean and standard deviation of all available values.
    The strategy assumes that the data is normally distributed.

    :param float lowerDeviationFactor: Detect anomalies if they are smaller than mean - lowerDeviationFactor * stdDev
    :param float upperDeviationFactor: Detect anomalies if they are bigger than mean + upperDeviationFactor * stdDev
    :param boolean includeInterval: Discerns whether or not the values inside the detection interval should
                be included in the calculation of the mean / stdDev.
    """

    def __init__(self, lowerDeviationFactor=3.0, upperDeviationFactor=3.0, includeInterval=False):
        self.lowerDeviationFactor = lowerDeviationFactor
        self.upperDeviationFactor = upperDeviationFactor
        self.includeInterval = includeInterval

    @property
    def _anomaly_jvm(self):
        return self._deequAnomalies.BatchNormalStrategy(
            self._jvm.scala.Option.apply(self.lowerDeviationFactor),
            self._jvm.scala.Option.apply(self.upperDeviationFactor),
            self.includeInterval,
        )


######################## HoltWinters  ########################


class MetricInterval(Enum):
    """Metric Interval is how often the metric of interest is computed (e.g. daily)."""

    Daily = "Daily"
    Monthly = "Monthly"

    def _get_java_object(self, jvm):
        if self == MetricInterval.Daily:
            return getattr(
                jvm.com.amazon.deequ.anomalydetection.seasonal.HoltWinters,
                "MetricInterval$",
            )().Daily()

        if self == MetricInterval.Monthly:
            return getattr(
                jvm.com.amazon.deequ.anomalydetection.seasonal.HoltWinters,
                "MetricInterval$",
            )().Monthly()
        raise ValueError("Invalid value for MetricInterval Enum")


class SeriesSeasonality(Enum):
    """
    SeriesSeasonality is the expected metric seasonality which defines the longest cycle in series. This is also
    referred to as periodicity.
    """

    Weekly = "Weekly"
    Yearly = "Yearly"

    def _get_java_object(self, jvm):
        if self == SeriesSeasonality.Weekly:
            return getattr(
                jvm.com.amazon.deequ.anomalydetection.seasonal.HoltWinters,
                "SeriesSeasonality$",
            )().Weekly()

        if self == SeriesSeasonality.Yearly:
            return getattr(
                jvm.com.amazon.deequ.anomalydetection.seasonal.HoltWinters,
                "SeriesSeasonality$",
            )().Yearly()
        raise ValueError("Invalid value for MetricInterval Enum")


class HoltWinters(_AnomalyObject):
    """
    Detects anomalies based on the additive Holt-Winters model.
    For example if a metric is produced daily and repeats itself every Monday,
    then the model should be created with a Daily metric interval and a Weekly seasonality parameter.
    To implement two cycles of data a minimum of 15 entries must be given for SeriesSeasonality.Weekly,
    MetricInterval.Daily, and 25 for SeriesSeasonality.Yearly, MetricInterval.Monthly.

    :param MetricInterval metricsInterval: How often a metric is available
    :param SeriesSeasonality seasonality: Cycle length (or periodicity) of the metric
    """

    def __init__(self, metricsInterval: MetricInterval, seasonality: SeriesSeasonality):
        self.metricsInterval = metricsInterval
        self.seasonality = seasonality

    @property
    def _anomaly_jvm(self):
        return self._deequAnomalies.seasonal.HoltWinters(
            self.metricsInterval._get_java_object(self._jvm), self.seasonality._get_java_object(self._jvm)
        )


######################## Anomaly Detector  ########################


class _AnomalyDetector(_AnomalyObject):

    """
    Detects anomalies in a time series give a strategy. This class is mainly responsible for
    the pre processing of the given time series.

    :param strategy:An implementation of AnomalyDetectionStrategy that needs preprocessed values.
    """

    def __init__(self, strategy):
        raise NotImplementedError("AnomalyDetector has not been implemented yet!")


class _DataPoint:
    def __init__(self, time, metricValue):
        raise NotImplementedError("DataPoint has not been implemented yet!")


######################## Detection Result  ########################


class _Anomaly:
    def __init__(self, value, confidence, detail):
        raise NotImplementedError("Anomaly has not been implemented yet!")



# -*- coding: utf-8 -*-
from functools import lru_cache
import os
import re


SPARK_TO_DEEQU_COORD_MAPPING = {
    "3.5": "com.amazon.deequ:deequ:2.0.7-spark-3.5",
    "3.3": "com.amazon.deequ:deequ:2.0.7-spark-3.3",
    "3.2": "com.amazon.deequ:deequ:2.0.7-spark-3.2",
    "3.1": "com.amazon.deequ:deequ:2.0.7-spark-3.1"
}


def _extract_major_minor_versions(full_version: str):
    major_minor_pattern = re.compile(r"(\d+\.\d+)\.*")
    match = re.match(major_minor_pattern, full_version)
    if match:
        return match.group(1)


@lru_cache(maxsize=None)
def _get_spark_version() -> str:
    try:
        spark_version = os.environ["SPARK_VERSION"]
    except KeyError:
        raise RuntimeError(f"SPARK_VERSION environment variable is required. Supported values are: {SPARK_TO_DEEQU_COORD_MAPPING.keys()}")

    return _extract_major_minor_versions(spark_version)


def _get_deequ_maven_config():
    spark_version = _get_spark_version()
    try:
        return SPARK_TO_DEEQU_COORD_MAPPING[spark_version[:3]]
    except KeyError:
        raise RuntimeError(
            f"Found incompatible Spark version {spark_version}; Use one of the Supported Spark versions for Deequ: {SPARK_TO_DEEQU_COORD_MAPPING.keys()}"
        )


SPARK_VERSION = _get_spark_version()
DEEQU_MAVEN_COORD = _get_deequ_maven_config()
IS_DEEQU_V1 = re.search("com\.amazon\.deequ\:deequ\:1.*", DEEQU_MAVEN_COORD) is not None



# -*- coding: utf-8 -*-
"""
Utility methods to help PyDeequ interface with Pandas
"""
import warnings
from typing import Union

from pandas import DataFrame as pandasDF
from pyspark.sql import DataFrame as sparkDF
from pyspark.sql import SparkSession
from pyspark.sql.types import (
    BooleanType,
    DateType,
    FloatType,
    IntegerType,
    LongType,
    StringType,
    StructField,
    StructType,
)


def ensure_pyspark_df(spark_session: SparkSession, df: Union[pandasDF, sparkDF]):
    """Method for checking dataframe type for each onData() call from a RunBuilder."""
    if not isinstance(df, sparkDF):
        warnings.warn(
            "WARNING: You passed in a Pandas DF, so we will be using our experimental utility to "
            "convert it to a PySpark DF."
        )
        df = PandasConverter.pandasDF_to_pysparkDF(spark_session, df)
    return df


class PandasConverter:
    """
    Pandas Functor to interface between Pandas and PyDeequ

    CONVERSIONS:
    Pandas DF DataType | PySpark DF DataType
    int32              | IntegerType()
    int64              | LongType()
    float64            | FloatType()
    bool               | BooleanType()
    datetime64[ns]     | DateType()
    object             | StringType()

    NOTE: LIMITED FUNCTIONALITY. Handle this prior and pass a Spark DF in if you want to customize
    - NaN columns get mapped to floats
    - Python lists & dicts get mapped to String
    - The following are unsupported automatic conversions: "TimestampType", "DecimalType", "DoubleType", "ByteType",
        "ShortType", "ArrayType", "MapType"
    """

    p2s_map = {
        "datetime64[ns]": DateType(),
        "int64": LongType(),
        "int32": IntegerType(),
        "float64": FloatType(),
        "bool": BooleanType(),
    }

    @classmethod
    def _define_structure(cls, string, format_type):
        typo = cls.p2s_map.get(str(format_type), StringType())
        return StructField(string, typo)

    @classmethod
    def pandasDF_to_pysparkDF(cls, spark_session: SparkSession, pandas_df: pandasDF):
        columns = list(pandas_df.columns)
        types = list(pandas_df.dtypes)
        struct_list = [cls._define_structure(column, typo) for column, typo in zip(columns, types)]

        p_schema = StructType(struct_list)
        return spark_session.createDataFrame(pandas_df, p_schema)

    @classmethod
    def pysparkDF_to_pandasDF(cls, spark_df: sparkDF):
        return spark_df.toPandas()



# -*- coding: utf-8 -*-
"""A collection of utility functions and classes for manipulating with scala objects anc classes through py4j
"""
from py4j.java_gateway import JavaObject


class PythonCallback:
    # https://www.py4j.org/advanced_topics.html#py4j-memory-model
    # https://stackoverflow.com/questions/50878834/py4j-error-while-obtaining-a-new-communication-channel-on-multithreaded-java
    def __init__(self, gateway):
        self.gateway = gateway
        # P4j will return false if the callback server is already started
        # https://github.com/bartdag/py4j/blob/master/py4j-python/src/py4j/java_gateway.py
        callback_server = self.gateway.get_callback_server()
        # TODO clean
        if callback_server is None:
            self.gateway.start_callback_server()
            print("Python Callback server started!")  # TODO Logging
        elif callback_server.is_shutdown:
            callback_server.close()
            self.gateway.restart_callback_server()
            # Have you tried turning it off and on again?
            # TODO why do we need to restart this every time?
            # TODO Will this break during chained function calls?
            print("PythonCallback server restarted!")


class ScalaFunction1(PythonCallback):
    """Implements scala.Function1 interface so we can pass lambda functions to Check
    https://www.scala-lang.org/api/current/scala/Function1.html"""

    def __init__(self, gateway, lambda_function):
        super().__init__(gateway)
        self.lambda_function = lambda_function

    def apply(self, arg):
        """Implements the apply function"""
        return self.lambda_function(arg)

    class Java:
        """scala.Function1: a function that takes one argument"""

        implements = ["scala.Function1"]


class ScalaFunction2(PythonCallback):
    """Implements scala.Function2 interface
    https://www.scala-lang.org/api/current/scala/Function2.html"""

    def __init__(self, gateway, lambda_function):
        super().__init__(gateway)
        self.lambda_function = lambda_function

    def apply(self, t1, t2):
        """Implements the apply function"""
        return self.lambda_function(t1, t2)

    class Java:
        """scala.Function2: a function that takes two arguments"""

        implements = ["scala.Function2"]


def get_or_else_none(scala_option):
    # TODO Checks
    if scala_option.isEmpty():
        return None
    return scala_option.get()


def to_scala_seq(jvm, iterable):
    """
    Helper method to take an iterable and turn it into a Scala sequence
    Args:
        jvm: Spark session's JVM
        iterable: your iterable
    Returns:
        Scala sequence
    """
    return jvm.scala.collection.JavaConversions.iterableAsScalaIterable(iterable).toSeq()


def to_scala_map(spark_session, d):
    """
    Convert a dict into a JVM Map.
    Args:
        spark_session: Spark session
        d: Python dictionary
    Returns:
        Scala map
    """
    return spark_session._jvm.PythonUtils.toScalaMap(d)


def scala_map_to_dict(jvm, scala_map):
    return dict(jvm.scala.collection.JavaConversions.mapAsJavaMap(scala_map))


def scala_map_to_java_map(jvm, scala_map):
    return jvm.scala.collection.JavaConversions.mapAsJavaMap(scala_map)


def java_list_to_python_list(java_list: str, datatype):
    # TODO: Improve
    # Currently turn the java object into a string and pass it through
    start = java_list.index("(")
    end = java_list.index(")")
    empty_val = "" if datatype == str else None
    return [datatype(i) if i != "" else empty_val for i in java_list[start + 1 : end].split(",")]


def scala_get_default_argument(java_object, argument_idx: int) -> JavaObject:
    return getattr(java_object, f"apply$default${argument_idx}")()



# -*- coding: utf-8 -*-
# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Placeholder docstrings"""
__version__ = "1.2.0"

from pyspark.sql import SparkSession

from pydeequ.analyzers import AnalysisRunner
from pydeequ.checks import Check, CheckLevel
from pydeequ.configs import DEEQU_MAVEN_COORD
from pydeequ.profiles import ColumnProfilerRunner

deequ_maven_coord = DEEQU_MAVEN_COORD
f2j_maven_coord = "net.sourceforge.f2j:arpack_combined_all"


class PyDeequSession:
    """
    For interacting with PyDeequ Modules at the "Runner" Level
    """

    def __init__(self, spark_session: SparkSession):
        self._spark_session = spark_session
        self._sc = spark_session.sparkContext
        self._jvm = spark_session._jvm

    def createColumnProfileRunner(self):
        return ColumnProfilerRunner(self._spark_session)

    def createAnalysisRunner(self):
        return AnalysisRunner(self._spark_session)

    def createCheck(self, level: CheckLevel, description: str, constraints=None):
        return Check(self._spark_session, level, description, constraints)



# -*- coding: utf-8 -*-
""" Profiles file for all the Profiles classes in Deequ"""
import json
from collections import namedtuple

from pyspark.sql import DataFrame, SparkSession
from pydeequ.analyzers import KLLParameters
from pydeequ.metrics import BucketDistribution
from pydeequ.pandas_utils import ensure_pyspark_df
from enum import Enum
from pydeequ.scala_utils import (
    get_or_else_none,
    java_list_to_python_list,
    scala_map_to_dict,
    scala_map_to_java_map,
    to_scala_map,
    to_scala_seq,
)

# ColumnProfilerRunner Classes
# TODO refactor documentation (https://devguide.python.org/documenting/)

DistributionValue = namedtuple("DistributionValue", "value count ratio")  # TODO: Revisit with Chris


class ColumnProfilerRunner:
    """
    Primary class for interacting with the profiles module.

    :param SparkSession spark_session: sparkSession
    """

    def __init__(self, spark_session: SparkSession):
        if not isinstance(spark_session, SparkSession):
            raise TypeError(f"Expected SparkSession object for spark_session, not {type(spark_session)}")

        self._spark_session = spark_session
        self._sc = spark_session.sparkContext
        self._jvm = spark_session._jvm
        self._ColumnProfilerRunner = self._jvm.com.amazon.deequ.profiles.ColumnProfilerRunner()

    def onData(self, df):
        """
        Starting point to construct a profile

        :param df: Tabular data on which the profiles module will use
        :return: The starting point to construct a profile
        """
        df = ensure_pyspark_df(self._spark_session, df)
        return ColumnProfilerRunBuilder(self._spark_session, df)

    def run(
        self,
        data,
        restrictToColumns,
        lowCardinalityHistogramThreshold,
        printStatusUpdates,
        cacheInputs,
        fileOutputOptions,
        metricsRepositoryOptions,
        kllParameters,
        predefinedTypes,
    ):
        """

        :param data:
        :param restrictToColumns:
        :param lowCardinalityHistogramThreshold:
        :param printStatusUpdates:
        :param cacheInputs:
        :param fileOutputOptions:
        :param metricsRepositoryOptions:
        :param kllParameters:
        :param predefinedTypes:
        :return:
        """
        raise NotImplementedError("Encapsulated")
        # TODO: preprocess parameters
        # return self._ColumnProfilerRunner.run(
        #     data,
        #     restrictToColumns,
        #     lowCardinalityHistogramThreshold,
        #     printStatusUpdates,
        #     cacheInputs,
        #     fileOutputOptions,
        #     metricsRepositoryOptions,
        #     kllParameters,
        #     predefinedTypes
        # )


# ColumnProfilerRunBuilder Classes


class ColumnProfilerRunBuilder:
    """
    Low level class for running profiling module

    :param SparkSession spark_session: sparkSession
    :param data: Tabular data which will be used to construct a profile.
    """

    def __init__(self, spark_session: SparkSession, data: DataFrame):

        if not isinstance(spark_session, SparkSession):
            raise TypeError(f"Expected SparkSession object for spark_session, not {type(spark_session)}")
        if not isinstance(data, DataFrame):
            raise TypeError(f"Expected DataFrame object for data, not {type(data)}")

        self._spark_session = spark_session
        self._sc = spark_session.sparkContext
        self._jvm = spark_session._jvm
        self._ColumnProfilerRunBuilder = self._jvm.com.amazon.deequ.profiles.ColumnProfilerRunBuilder(data._jdf)

    def run(self):
        """
        A method that runs a profile check on the data to obtain a ColumnProfiles class

        :return: A ColumnProfiles result
        """
        run = self._ColumnProfilerRunBuilder.run()
        return ColumnProfilesBuilder(self._spark_session)._columnProfilesFromColumnRunBuilderRun(run)

    def printStatusUpdates(self, print_status_updates: bool):
        """
        Print status updates between passes

        :param bool print_status_updates: Whether to print status updates
        :return: Printed status
        """
        self._ColumnProfilerRunBuilder.printStatusUpdates(print_status_updates)
        return self

    def cacheInputs(self, cache_inputs: bool):
        """
        Cache the inputs

        :param bool cache_inputs: Whether to print status updates
        :return: Cache inputs
        """
        self._ColumnProfilerRunBuilder.cacheInputs(cache_inputs)
        return self

    def withLowCardinalityHistogramThreshold(self, low_cardinality_histogram_threshold: int):
        """
        Set the thresholds of value until it is expensive to calculate the histograms

        :param int low_cardinality_histogram_threshold: The designated threshold
        :return: a set threshold
        """
        self._ColumnProfilerRunBuilder.withLowCardinalityHistogramThreshold(low_cardinality_histogram_threshold)
        return self

    def restrictToColumns(self, restrict_to_columns: list):
        """
        Can be used to specify a subset of columns to look out

        :param list restrict_to_columns: Specified columns
        :return: A subset of columns to look at
        """
        self._ColumnProfilerRunBuilder.restrictToColumns(to_scala_seq(self._jvm, restrict_to_columns))
        return self

    def withKLLProfiling(self):
        """
        Enable KLL Sketches profiling on Numerical columns, disabled by default.

        :return: Enable KLL Sketches profiling on Numerical columns, disabled by default.
        """
        self._ColumnProfilerRunBuilder.withKLLProfiling()
        return self

    def setKLLParameters(self, kllParameters: KLLParameters):
        """
        Set kllParameters

        :param KLLParameters kllParameters: kllParameters(sketchSize, shrinkingFactor, numberOfBuckets)
        """
        self._ColumnProfilerRunBuilder.setKLLParameters(self._jvm.scala.Option.apply(kllParameters._param))
        return self

    def setPredefinedTypes(self, dataTypesDict: dict):
        """
        Set predefined data types for each column (e.g. baseline)

        :param dict{"columnName": DataTypeInstance} dataTypes: dataType map for baseline columns.
        :return: Baseline for each column. I.E. returns the dataType label to the desired DataTypeInstance
        """
        dataType_scala_map = {}
        for key, value in dataTypesDict.items():
            val = value._create_java_object(self._jvm)
            dataType_scala_map[key] = val
        self._ColumnProfilerRunBuilder.setPredefinedTypes(to_scala_map(self._spark_session, dataType_scala_map))
        return self

    def useRepository(self, repository):
        """
        Set a metrics repository associated with the current data to enable features like reusing
        previously computed results and storing the results of the current run.

        :param repository:A metrics repository to store and load results associated with the run
        :return: Sets a metrics repository with the current data to use features on
        """
        self._ColumnProfilerRunBuilder = self._ColumnProfilerRunBuilder.useRepository(repository.repository)
        return self

    def saveOrAppendResult(self, resultKey):
        """
        A shortcut to save the results of the run or append them to existing results in the metrics repository

        :param resultKey: The result key to identify the current run
        :return: A saved results of the run in the metrics repository
        """

        self._ColumnProfilerRunBuilder.saveOrAppendResult(resultKey.resultKey)
        return self

    def useSparkSession(self, sparkSession):
        """
        Use a sparkSession to conveniently create output files

        :param SparkSession sparkSession: sparkSession
        :return: A sparksession to create output files
        """
        # TODO


class ColumnProfilesBuilder:
    def __init__(self, spark_session: SparkSession):
        """
        The results returned from the columnProfilerRunner

        :param SparkSession spark_session: sparkSession
        """
        if not isinstance(spark_session, SparkSession):
            raise TypeError(f"Expected SparkSession object, not {type(spark_session)}")

        self._spark_session = spark_session
        self._sc = spark_session.sparkContext
        self._jvm = spark_session._jvm
        self._profiles = []
        self.columnProfileClasses = {
            "StandardColumnProfile": StandardColumnProfile,
            "StringColumnProfile": StandardColumnProfile,
            "NumericColumnProfile": NumericColumnProfile,

        }

    def _columnProfilesFromColumnRunBuilderRun(self, run):
        """
        Produces a Java profile based on the designated column

        :param run: columnProfilerRunner result
        :return: a setter for columnProfilerRunner result
        """
        self._run_result = run
        profile_map = self._jvm.scala.collection.JavaConversions.mapAsJavaMap(run.profiles())  # TODO from ScalaUtils
        self._profiles = {column: self._columnProfileBuilder(column, profile_map[column]) for column in profile_map}
        return self

    @property
    def profiles(self):
        """
        A getter for profiles

        :return: a getter for profiles
        """
        return self._profiles

    def _columnProfileBuilder(self, column, java_column_profile):
        """Factory function for ColumnProfile
        Returns a Java profile based on the designated column

        :param column: The column to run a profile on
        :param java_column_profile: The profile mapped as a Java map
        """
        return self.columnProfileClasses[java_column_profile.getClass().getSimpleName()](
            self._spark_session, column, java_column_profile
        )


class ColumnProfile:
    """Factory class for Standard and Numeric Column Profiles
    The class for getting the Standard and Numeric Column profiles of the data.

    :param SparkSession spark_session: sparkSession
    :param column: designated column to run a profile on
    :param java_column_profile: The profile mapped as a Java map
    """

    def __init__(self, spark_session: SparkSession, column, java_column_profile):
        if not isinstance(spark_session, SparkSession):
            raise TypeError(f"Expected SparkSession object for spark_session, not {type(spark_session)}")
        self._spark_session = spark_session
        self._sc = spark_session.sparkContext
        self._jvm = spark_session._jvm
        self._java_column_profile = java_column_profile
        self._column = column
        self._completeness = java_column_profile.completeness()
        self._approximateNumDistinctValues = java_column_profile.approximateNumDistinctValues()
        self._dataType = java_column_profile.dataType()
        self._typeCounts = scala_map_to_dict(self._jvm, java_column_profile.typeCounts())
        self._isDataTypeInferred = java_column_profile.isDataTypeInferred() == "true"
        if get_or_else_none(self._java_column_profile.histogram()):
            self._histogram = [
                DistributionValue(k, v.absolute(), v.ratio())
                for k, v in scala_map_to_java_map(
                    self._jvm, self._java_column_profile.histogram().get().values()
                ).items()
            ]
        else:
            self._histogram = None

    @property
    def column(self):
        """
        Getter for the current column name in ColumnProfile

        :return:  gets the column name in the Column Profile
        """
        return self._column

    @property
    def completeness(self):
        """ "
        Getter that returns the completeness of data in the column

        :return:  gets the calculated completeness of data in the column
        """
        return self._completeness

    @property
    def approximateNumDistinctValues(self):
        """
        Getter that returns the amount of distinct values in the column

        :return: gets the number of distinct values in the column
        """
        return self._approximateNumDistinctValues

    @property
    def dataType(self):
        """
        Getter that returns the datatype of the column

        :return: gets the datatype of the column
        """
        return str(self._dataType)

    @property
    def isDataTypeInferred(self):
        """
        Getter that returns a boolean of whether the Data Type of the column was inferred

        :return: gets the isDataTypeInferred of the column
        """
        return self._isDataTypeInferred

    @property
    def typeCounts(self):
        """
        A getter for the number of values for each datatype in the column

        :return: gets the number of values for each datatype
        """
        return self._typeCounts

    @property
    def histogram(self):
        """
        A getter for the full value distribution of the column

        :return: gets the histogram of a column
        """
        return self._histogram


class StandardColumnProfile(ColumnProfile):
    """
    Standard Column Profile class

    :param SparkSession spark_session: sparkSession
    :param column: the designated column of which the profile is run on
    :param java_column_profile: The profile mapped as a Java map
    """

    def __init__(self, spark_session: SparkSession, column, java_column_profile):
        super().__init__(spark_session, column, java_column_profile)
        self.all = {
            "completeness": self.completeness,
            "approximateNumDistinctValues": self.approximateNumDistinctValues,
            "dataType": self.dataType,
            "isDataTypeInferred": self.isDataTypeInferred,
            "typeCounts": self.typeCounts,
            "histogram": self.histogram,
        }

    def __str__(self):
        """
        A JSON of the standard profiles for each column

        :return: A JSON of the standard profiles
        """
        return f"StandardProfiles for column: {self.column}: {json.dumps(self.all, indent=4)}"


class NumericColumnProfile(ColumnProfile):
    """
    Numeric Column Profile class

    :param SparkSession spark_session: sparkSession
    :param column: the designated column of which the profile is run on
    :param java_column_profile: The profile mapped as a Java map
    """

    def __init__(self, spark_session: SparkSession, column, java_column_profile):
        super().__init__(spark_session, column, java_column_profile)
        # TODO: self.numRecords = java_column_profile.numRecords()
        self._kll = (
            BucketDistribution(spark_session, java_column_profile.kll().get())
            if get_or_else_none(java_column_profile.kll())
            else None
        )
        self._mean = get_or_else_none(java_column_profile.mean())
        self._maximum = get_or_else_none(java_column_profile.maximum())
        self._minimum = get_or_else_none(java_column_profile.minimum())
        self._sum = get_or_else_none(java_column_profile.sum())
        self._stdDev = get_or_else_none(java_column_profile.stdDev())
        self._approxPercentiles = (
            java_list_to_python_list(str(get_or_else_none(java_column_profile.approxPercentiles())), float)
            if get_or_else_none(java_column_profile.approxPercentiles())
            else []
        )
        self.all = {
            "completeness": self.completeness,
            "approximateNumDistinctValues": self.approximateNumDistinctValues,
            "dataType": self.dataType,
            "isDataTypeInferred": self.isDataTypeInferred,
            "typeCounts": self.typeCounts,
            "histogram": self.histogram,
            "kll": str(self._kll),
            "mean": self._mean,
            "maximum": self._maximum,
            "minimum": self._minimum,
            "sum": self._sum,
            "stdDev": self._stdDev,
            "approxPercentiles": self._approxPercentiles,
        }

    def __str__(self):
        """
        A JSON of the numerical profiles for each column

        :return: A JSON of the numerical profiles

        """
        return f"NumericProfiles for column: {self.column}: {json.dumps(self.all, indent=4)}"

    @property
    def kll(self):
        """
        A getter for the kll value of a numeric column
        :return: gets the kll value of a numeric column
        """
        return self._kll

    @property
    def mean(self):
        """
        A getter for the calculated mean of the numeric column

        :return: gets the mean of the column
        """
        return self._mean

    @property
    def maximum(self):
        """
        A getter for the maximum value of the numeric column

        :return: gets the maximum value of the column
        """
        return self._maximum

    @property
    def minimum(self):
        """
        A getter for the minimum value of the numeric column

        :return: gets the minimum value of the column
        """
        return self._minimum

    @property
    def sum(self):
        """
        A getter for the sum of the numeric column

        :return: gets the sum value of the column
        """
        return self._sum

    @property
    def stdDev(self):
        """
        A getter for the standard deviation  of the numeric column

        :return: gets the standard deviation of the column
        """
        return self._stdDev

    @property
    def approxPercentiles(self):
        """
        A getter for the approximate percentiles  of the numeric column

        :return: gets the approximate percentiles of the column
        """
        return self._approxPercentiles




# -*- coding: utf-8 -*-
"""
Metrics file for the various metrics that Deequ handles
"""
from pyspark.sql import SparkSession


class BucketValue:
    def __init__(self, bucket_value):
        self._bucketValue = bucket_value

    def __str__(self):
        return str(self._bucketValue)

    @property
    def lowValue(self):
        return float(self._bucketValue.lowValue())

    @property
    def highValue(self):
        return float(self._bucketValue.highValue())

    @property
    def count(self):
        return int(self._bucketValue.count())


class BucketDistribution:
    """
    Bucket Distribution for KLL
    """

    def __init__(self, spark_session: SparkSession, kllResult):
        self._spark_session = spark_session
        self._kllResult = kllResult
        self._buckets = [
            BucketValue(self._kllResult.buckets().apply(b)) for b in range(self._kllResult.buckets().size())
        ]
        self._parameters = [
            float(self._kllResult.parameters().apply(p)) for p in range(self._kllResult.parameters().size())
        ]
        self._data = [list(d) for d in self._kllResult.data()]

    def __str__(self):
        # TODO: Reformat
        return str(self._kllResult)

    @property
    def buckets(self):
        return self._buckets

    @property
    def parameters(self):
        return self._parameters

    @property
    def data(self):
        return self._data

    def computePercentiles(self):
        """
        Compute the percentiles.
        """
        return list(self._kllResult.computePercentiles())

    def apply(self, key: int):
        """
        Get relevant bucketValue with index of bucket.

        :param int key: index of bucket
        :return BucketValue: The metrics for the bucket
        """
        return BucketValue(self._kllResult.apply(key))

    @property
    def argmax(self):
        """
        Find the index of bucket which contains the most items.

        :return: The index of bucket which contains the most items.
        """
        return self._kllResult.argmax()



# -*- coding: utf-8 -*-
from enum import Enum

from py4j.protocol import Py4JError
from pyspark.sql import SparkSession

from pydeequ.check_functions import is_one
from pydeequ.scala_utils import ScalaFunction1, to_scala_seq
from pydeequ.configs import SPARK_VERSION

# TODO implement custom assertions
# TODO implement all methods without outside class dependencies
# TODO Integration with Constraints


class CheckLevel(Enum):
    Error = "Error"
    Warning = "Warning"

    def _get_java_object(self, jvm):
        if self == CheckLevel.Error:
            return jvm.com.amazon.deequ.checks.CheckLevel.Error()
        if self == CheckLevel.Warning:
            return jvm.com.amazon.deequ.checks.CheckLevel.Warning()
        raise ValueError("Invalid value for CheckLevel Enum")


class CheckStatus(Enum):
    Success = "Success"
    Warning = "Warning"
    Error = "Error"

    @classmethod
    def _create_from_java_object(cls, java_object):
        check_status_class = java_object._jvm.com.amazon.deequ.checks.CheckStatus
        if java_object.equals(check_status_class.Success()):
            return CheckStatus.Success
        if java_object.equals(check_status_class.Warning()):
            return CheckStatus.Warning
        if java_object.equals(check_status_class.Error()):
            return CheckStatus.Error
        raise ValueError(f"{java_object} is not a valid CheckStatus Object")


class ConstrainableDataTypes(Enum):
    Null = "Null"
    Fractional = "Fractional"
    Integral = "Integral"
    Boolean = "Boolean"
    String = "String"
    Numeric = "Numeric"

    def _get_java_object(self, jvm):
        if self == ConstrainableDataTypes.Null:
            return jvm.com.amazon.deequ.constraints.ConstrainableDataTypes.Null()
        if self == ConstrainableDataTypes.Fractional:
            return jvm.com.amazon.deequ.constraints.ConstrainableDataTypes.Fractional()
        if self == ConstrainableDataTypes.Integral:
            return jvm.com.amazon.deequ.constraints.ConstrainableDataTypes.Integral()
        if self == ConstrainableDataTypes.Boolean:
            return jvm.com.amazon.deequ.constraints.ConstrainableDataTypes.Boolean()
        if self == ConstrainableDataTypes.String:
            return jvm.com.amazon.deequ.constraints.ConstrainableDataTypes.String()
        if self == ConstrainableDataTypes.Numeric:
            return jvm.com.amazon.deequ.constraints.ConstrainableDataTypes.Numeric()
        raise ValueError("Invalid value for ConstrainableDataType Enum")


class CheckResult:
    pass


class Check:
    """
    A class representing a list of constraints that can be applied to a given
     [[org.apache.spark.sql.DataFrame]]. In order to run the checks, use the `run` method. You can
     also use VerificationSuite.run to run your checks along with other Checks and Analysis objects.
     When run with VerificationSuite, Analyzers required by multiple checks/analysis blocks is
     optimized to run once.
    """

    def __init__(self, spark_session: SparkSession, level: CheckLevel, description: str, constraints: list = None):
        """ "

        :param SparkSession spark_session: SparkSession
        :param CheckLevel level: Assertion level of the check group. If any of the constraints fail this level is used for
                the status of the check.
        :param str description: The name describes the check block. Generally it will be used to show in the logs.
        :param list constraints: The constraints to apply when this check is run. New ones can be added
                and will return a new object
        """
        self._spark_session = spark_session
        self._sc = spark_session.sparkContext
        self._jvm = spark_session._jvm
        self.level = level
        self._java_level = self.level._get_java_object(self._jvm)
        self._check_java_class = self._jvm.com.amazon.deequ.checks.Check
        self.description = description
        self._Check = self._check_java_class(
            self._java_level, self.description, getattr(self._check_java_class, "apply$default$3")()
        )
        self.constraints = constraints or []
        for constraint in self.constraints:
            self.addConstraint(constraint)

    def addConstraints(self, constraints: list):
        self.constraints.extend(constraints)
        for constraint in constraints:
            self._Check = constraint._Check

    def addConstraint(self, constraint):
        """
        Returns a new Check object with the given constraints added to the constraints list.
        :param Constraint constraint: new constraint to be added.
        :return: new Check object
        """
        self.constraints.append(constraint)
        self._Check = constraint._Check

    def where(self, filter: str):
        try:
            self._Check = self._Check.where(filter)
        except Py4JError:
            raise TypeError(f"Method doesn't exist in {self._Check.getClass()}, class has to be filterable")
        return self

    def addFilterableContstraint(self, creationFunc):
        """Adds a constraint that can subsequently be replaced with a filtered version
        :param creationFunc:
        :return:
        """
        raise NotImplementedError("Private factory method for other check methods")

    def hasSize(self, assertion, hint=None):
        """
        Creates a constraint that calculates the data frame size and runs the assertion on it.

        :param lambda assertion: Refers to a data frame size. The given function can include comparisons
                and conjunction or disjunction statements.
        :param str hint: A hint that states why a constraint could have failed.
        :return: hasSize self: A Check.scala object that implements the assertion on the column.

        """
        assertion_func = ScalaFunction1(self._spark_session.sparkContext._gateway, assertion)
        hint = self._jvm.scala.Option.apply(hint)
        self._Check = self._Check.hasSize(assertion_func, hint)  # getattr(self._Check, "hasSize$default$2"))
        return self

    def isComplete(self, column, hint=None):
        """Creates a constraint that asserts on a column completion.

        :param str column: Column in Data Frame to run the assertion on.
        :param str hint: A hint that discerns why a constraint could have failed.
        :return: isComplete self:A Check.scala object that asserts on a column completion.
        """
        hint = self._jvm.scala.Option.apply(hint)
        self._Check = self._Check.isComplete(column, hint, self._jvm.scala.Option.apply(None))
        return self

    def hasCompleteness(self, column, assertion, hint=None):
        """
        Creates a constraint that asserts column completion.
        Uses the given history selection strategy to retrieve historical completeness values on this
        column from the history provider.

        :param str column: Column in Data Frame to run the assertion on.
        :param lambda assertion: A function that accepts an int or float parameter.
        :param str hint: A hint that states why a constraint could have failed.
        :return: hasCompleteness self: A Check object that implements the assertion.
        """
        assertion_func = ScalaFunction1(self._spark_session.sparkContext._gateway, assertion)
        hint = self._jvm.scala.Option.apply(hint)
        self._Check = self._Check.hasCompleteness(column, assertion_func, hint, self._jvm.scala.Option.apply(None))
        return self

    def areComplete(self, columns, hint=None):
        """Creates a constraint that asserts completion in combined set of columns.

        :param list[str] columns: Columns in Data Frame to run the assertion on.
        :param str hint: A hint that states why a constraint could have failed.
        :return: areComplete self: A Check.scala object that asserts completion in the columns.
        """
        hint = self._jvm.scala.Option.apply(hint)
        columns_seq = to_scala_seq(self._jvm, columns)
        self._Check = self._Check.areComplete(columns_seq, hint)
        return self

    def haveCompleteness(self, columns, assertion, hint=None):
        """Creates a constraint that asserts on completed rows in a combined set of columns.

        :param list[str] columns: Columns in Data Frame to run the assertion on.
        :param lambda assertion: A function that accepts an int or float parameter.
        :param str hint: A hint that states why a constraint could have failed.
        :return: haveCompleteness self: A Check.scala object that implements the assertion on the columns.
        """
        columns_seq = to_scala_seq(self._jvm, columns)
        assertion_func = ScalaFunction1(self._spark_session.sparkContext._gateway, assertion)
        hint = self._jvm.scala.Option.apply(hint)
        self._Check = self._Check.haveCompleteness(columns_seq, assertion_func, hint)
        return self

    def areAnyComplete(self, columns, hint=None):
        """Creates a constraint that asserts any completion in the combined set of columns.

        :param list[str] columns: Columns in Data Frame to run the assertion on.
        :param str hint: A hint that states why a constraint could have failed.
        :return: areAnyComplete self: A Check.scala object that asserts completion in the columns.
        """
        hint = self._jvm.scala.Option.apply(hint)
        columns_seq = to_scala_seq(self._jvm, columns)
        self._Check = self._Check.areAnyComplete(columns_seq, hint)
        return self

    def haveAnyCompleteness(self, columns, assertion, hint=None):
        """Creates a constraint that asserts on any completion in the combined set of columns.

        :param list[str] columns: Columns in Data Frame to run the assertion on.
        :param lambda assertion: A function that accepts an int or float parameter.
        :param str hint: A hint that states why a constraint could have failed.
        :return: haveAnyCompleteness self: A Check.scala object that asserts completion in the columns.
        """
        columns_seq = to_scala_seq(self._jvm, columns)
        assertion_func = ScalaFunction1(self._spark_session.sparkContext._gateway, assertion)
        hint = self._jvm.scala.Option.apply(hint)
        self._Check = self._Check.haveAnyCompleteness(columns_seq, assertion_func, hint)
        return self

    def isUnique(self, column, hint=None):
        """
        Creates a constraint that asserts on a column uniqueness

        :param list[str] columns: Columns in Data Frame to run the assertion on.
        :param str hint: A hint that states why a constraint could have failed.
        :return: isUnique self: A Check.scala object that asserts uniqueness in the column.
        """
        hint = self._jvm.scala.Option.apply(hint)
        self._Check = self._Check.isUnique(column, hint, self._jvm.scala.Option.apply(None))
        return self

    def isPrimaryKey(self, column, *columns, hint=None):
        """
        Creates a constraint that asserts on a column(s) primary key characteristics.
        Currently only checks uniqueness, but reserved for primary key checks if there is another
        assertion to run on primary key columns.

        #  how does column and columns differ
        :param str column: Column in Data Frame to run the assertion on.
        :param str hint: A hint that states why a constraint could have failed.
        :param list[str] columns: Columns to run the assertion on.
        :return: isPrimaryKey self: A Check.scala object that asserts completion in the columns.
        """
        hint = self._jvm.scala.Option.apply(hint)
        print(f"Unsolved integration: {hint}")
        raise NotImplementedError("Unsolved integration of Python tuple => varArgs")

    def hasUniqueness(self, columns, assertion, hint=None):
        """
        Creates a constraint that asserts any uniqueness in a single or combined set of key columns.
        Uniqueness is the fraction of unique values of a column(s)  values that occur exactly once.

        :param list[str] columns: Column(s) in Data Frame to run the assertion on.
        :param lambda assertion: A function that accepts an int or float parameter.
        :param str hint: A hint that states why a constraint could have failed.
        :return: hasUniqueness self: A Check object that asserts uniqueness in the columns.
        """
        assertion_func = ScalaFunction1(self._spark_session.sparkContext._gateway, assertion)
        columns_seq = to_scala_seq(self._jvm, columns)
        hint = self._jvm.scala.Option.apply(hint)
        self._Check = self._Check.hasUniqueness(columns_seq, assertion_func, hint)
        return self

    def hasDistinctness(self, columns, assertion, hint=None):
        """
        Creates a constraint on the distinctness in a single or combined set of key columns.
        Distinctness is the fraction of distinct values of a column(s).

        :param list[str] columns: Column(s) in Data Frame to run the assertion on.
        :param lambda assertion : A function that accepts an int or float parameter.
        :param str hint: A hint that states why a constraint could have failed.
        :return: hasDistinctness self: A Check object that asserts distinctness in the columns.
        """
        assertion_func = ScalaFunction1(self._spark_session.sparkContext._gateway, assertion)
        hint = self._jvm.scala.Option.apply(hint)
        columns_seq = to_scala_seq(self._jvm, columns)
        self._Check = self._Check.hasDistinctness(columns_seq, assertion_func, hint)
        return self

    def hasUniqueValueRatio(self, columns, assertion, hint=None):
        """
        Creates a constraint on the unique value ratio in a single or combined set of key columns.

        :param list[str] columns: Column(s) in Data Frame to run the assertion on.
        :param lambda assertion: A function that accepts an int or float parameter.
        :param str hint: A hint that states why a constraint could have failed.
        :return: hasUniqueValueRatio self: A Check object that asserts the unique value ratio in the columns.
        """
        assertion_func = ScalaFunction1(self._spark_session.sparkContext._gateway, assertion)
        hint = self._jvm.scala.Option.apply(hint)
        columns_seq = to_scala_seq(self._jvm, columns)
        self._Check = self._Check.hasUniqueValueRatio(columns_seq, assertion_func, hint, self._jvm.scala.Option.apply(None))
        return self

    def hasNumberOfDistinctValues(self, column, assertion, binningUdf, maxBins, hint=None):
        """Creates a constraint that asserts on the number of distinct values a column has.

        :param str column: Column in Data Frame to run the assertion on.
        :param lambda assertion: A function that accepts an int or float parameter.
        :param lambda binningUDF: An optional binning function.
        :param int maxBins: Histogram details is only provided for N column values with top counts. MaxBins sets the N.
        :param str hint: A hint that states why a constraint could have failed.
        :return: hasNumberOfDistinctValues self: A Check object that asserts distinctness in the column.
        """
        assertion_func = ScalaFunction1(self._spark_session.sparkContext._gateway, assertion)
        hint = self._jvm.scala.Option.apply(hint)
        self._Check = self._Check.hasNumberOfDistinctValues(column, assertion_func, binningUdf, maxBins, hint)
        return self

    def hasHistogramValues(self, column, assertion, binningUdf, maxBins, hint=None):
        """Creates a constraint that asserts on column's value distribution.

        :param str column: Column in Data Frame to run the assertion on.
        :param lambda assertion: A function that accepts an int or float parameter as a distribution input parameter.
        :param str binningUDF: An optional binning function.
        :param str maxBins: Histogram details is only provided for N column values with top counts. MaxBins sets the N.
        :param str hint: A hint that states why a constraint could have failed.
        :return: hasHistogramValues self: A Check object that asserts the column's value distribution in the column.
        """
        assertion_func = ScalaFunction1(self._spark_session.sparkContext._gateway, assertion)
        hint = self._jvm.scala.Option.apply(hint)
        self._Check = self._Check.hasHistogramValues(column, assertion_func, binningUdf, maxBins, hint)
        return self

    def kllSketchSatisfies(self, column, assertion, kllParameters=None, hint=None):
        """Creates a constraint that asserts on column's sketch size.

        :param str column: Column in Data Frame to run the assertion on.
        :param Lambda(BucketDistribution) assertion: A function that accepts an int or float
                parameter as a distribution input parameter.
        :param KLLParameters kllParameters: Parameters of KLL sketch
        :param str hint: A hint that states why a constraint could have failed.
        :return: kllSketchSatisfies self: A Check object that asserts the column's sketch size in the column.
        """
        assertion_func = ScalaFunction1(self._spark_session.sparkContext._gateway, assertion)
        hint = self._jvm.scala.Option.apply(hint)
        params = self._jvm.scala.Option.apply(kllParameters._param if kllParameters else None)
        self._Check = self._Check.kllSketchSatisfies(column, assertion_func, params, hint)
        return self

    def _isNewestPointNonAnomalous(self):
        """Creates a constraint that runs AnomalyDetection on the new value

        :param MetricsRepostiroy metricsRepository  (MetricsRepository): A metrics repository to get previous results.
        :param AnomalyDetectionStrategy anomalyDetectionStrategy (AnomalyDetectionStrategy ): The anomaly detection strategy.
        :param Analzyer analyzer: The analyzer for the metric to run anomaly detection on.
        :param withTagValues: Can contain a Map with tag names and the corresponding values to filter for.
        :param beforeDate: The maximum dateTime of previous AnalysisResults to use for the Anomaly Detection.
        :param afterDate: The minimum dateTime of previous AnalysisResults to use for the Anomaly Detection.
        :param str hint: A hint that states why a constraint could have failed.
        :return: isNewestPointAnomalous self: A Check object that asserts the column's sketch size in the columns.
        """
        raise NotImplementedError("private method in scala, will need anomaly detection implemented beforehand")

    def hasEntropy(self, column, assertion, hint=None):
        """
        Creates a constraint that asserts on a column entropy. Entropy is a measure of the level of information
        contained in a message.

        :param str column: Column in Data Frame to run the assertion on.
        :param lambda assertion: A function that accepts an int or float parameter.
        :param str hint: A hint that states why a constraint could have failed.
        :return: hasEntropy self: A Check object that asserts the entropy in the column.
        """
        assertion_func = ScalaFunction1(self._spark_session.sparkContext._gateway, assertion)
        hint = self._jvm.scala.Option.apply(hint)
        self._Check = self._Check.hasEntropy(column, assertion_func, hint)
        return self

    def hasMutualInformation(self, columnA, columnB, assertion, hint=None):
        """
        Creates a constraint that asserts on a mutual information between two columns. Mutual Information describes how
        much information about one column can be inferred from another.

        :param str columnA: First column in Data Frame which calculates the mutual information.
        :param str columnB: Second column in Data Frame which calculates the mutual information.
        :param lambda assertion: A function that accepts an int or float parameter.
        :param str hint: A hint that states why a constraint could have failed.
        :return: hasMutualInformation self: A Check object that asserts the mutual information in the columns.
        """
        assertion_func = ScalaFunction1(self._spark_session.sparkContext._gateway, assertion)
        hint = self._jvm.scala.Option.apply(hint)
        self._Check = self._Check.hasMutualInformation(columnA, columnB, assertion_func, hint)
        return self

    def hasApproxQuantile(self, column, quantile, assertion, hint=None):
        """
        Creates a constraint that asserts on an approximated quantile

        :param str column: Column in Data Frame to run the assertion on
        :param float quantile: Quantile to run the assertion on.
        :param lambda assertion: A function that accepts the computed quantile as an int or float parameter.
        :param str hint: A hint that states why a constraint could have failed.
        :return: hasApproxQuantile self: A Check object that asserts the approximated quantile in the column.
        """
        assertion_func = ScalaFunction1(self._spark_session.sparkContext._gateway, assertion)
        hint = self._jvm.scala.Option.apply(hint)
        self._Check = self._Check.hasApproxQuantile(column, float(quantile), assertion_func, hint)
        return self

    def hasMinLength(self, column, assertion, hint=None):
        """
        Creates a constraint that asserts on the minimum length of a string datatype column.

        :param str column: Column in Data Frame to run the assertion on. The column is expected to be a string type.
        :param lambda assertion: A function which accepts the int or float parameter
                that discerns the minimum length of the string.
        :param str hint: A hint that states why a constraint could have failed.
        :return: hasMinLength self: A Check object that asserts minLength of the column.
        """
        assertion_func = ScalaFunction1(self._spark_session.sparkContext._gateway, assertion)
        hint = self._jvm.scala.Option.apply(hint)
        self._Check = self._Check.hasMinLength(column, assertion_func, hint, self._jvm.scala.Option.apply(None))
        return self

    def hasMaxLength(self, column, assertion, hint=None):
        """
        Creates a constraint that asserts on the maximum length of a string datatype column

        :param str column: Column in Data Frame to run the assertion on. The column is expected to be a string type.
        :param lanmbda assertion: A function which accepts an int or float parameter
                that discerns the maximum length of the string.
        :param str hint: A hint that states why a constraint could have failed.
        :return: hasMaxLength self : A Check object that asserts maxLength of the column.
        """
        assertion_func = ScalaFunction1(self._spark_session.sparkContext._gateway, assertion)
        hint = self._jvm.scala.Option.apply(hint)
        self._Check = self._Check.hasMaxLength(column, assertion_func, hint, self._jvm.scala.Option.apply(None))
        return self

    def hasMin(self, column, assertion, hint=None):
        """
        Creates a constraint that asserts on the minimum of a column. The column is contains either a long,
        int or float datatype.

        :param str column: Column in Data Frame to run the assertion on. The column is expected to be an int, long or float type.
        :param lambda assertion: A function which accepts an int or float parameter
                that discerns the minumum value of the column.
        :param str hint: A hint that states why a constraint could have failed.
        :return: hasMaxLength self: A Check object that asserts the minumum of the column.
        """
        assertion_func = ScalaFunction1(self._spark_session.sparkContext._gateway, assertion)
        hint = self._jvm.scala.Option.apply(hint)
        self._Check = self._Check.hasMin(column, assertion_func, hint, self._jvm.scala.Option.apply(None))
        return self

    def hasMax(self, column, assertion, hint=None):
        """
        Creates a constraint that asserts on the maximum of the column. The column contains either a long,
        int or float datatype.

        :param str column: Column in Data Frame to run the assertion on. The column is expected to be an int, long or float type.
        :param lambda assertion: A function which accepts an int or float parameter
                that discerns the maximum value of the column.
        :param str hint: A hint that states why a constraint could have failed.
        :return: hasMax self: A Check object that asserts maximum of the column.
        """
        assertion_func = ScalaFunction1(self._spark_session.sparkContext._gateway, assertion)
        hint = self._jvm.scala.Option.apply(hint)
        self._Check = self._Check.hasMax(column, assertion_func, hint, self._jvm.scala.Option.apply(None))
        return self

    def hasMean(self, column, assertion, hint=None):
        """
        Creates a constraint that asserts on the mean of the column

        :param str column: Column in Data Frame to run the assertion on.
        :param lambda assertion: A function with an int or float parameter. The parameter is the mean.
        :param str hint: A hint that states why a constraint could have failed.
        :return: hasMean self: A Check object that asserts the mean of the column.
        """
        assertion_func = ScalaFunction1(self._spark_session.sparkContext._gateway, assertion)
        hint = self._jvm.scala.Option.apply(hint)
        self._Check = self._Check.hasMean(column, assertion_func, hint)
        return self

    def hasSum(self, column, assertion, hint=None):
        """
        Creates a constraint that asserts on the sum of the column

        :param str column: Column in Data Frame to run the assertion on.
        :param lambda assertion: A function with an int or float parameter. The parameter is the sum.
        :param str hint: A hint that states why a constraint could have failed.
        :return: hasMean self: A Check object that asserts the sum of the column.
        """
        assertion_func = ScalaFunction1(self._spark_session.sparkContext._gateway, assertion)
        hint = self._jvm.scala.Option.apply(hint)
        self._Check = self._Check.hasSum(column, assertion_func, hint)
        return self

    def hasStandardDeviation(self, column, assertion, hint=None):
        """
        Creates a constraint that asserts on the standard deviation of the column

        :param str column: Column in Data Frame to run the assertion on.
        :param lambda assertion: A function with an int or float parameter. The parameter is the standard deviation.
        :param str hint: A hint that states why a constraint could have failed.
        :return: hasMean self: A Check object that asserts the std deviation of the column.
        """
        assertion_func = ScalaFunction1(self._spark_session.sparkContext._gateway, assertion)
        hint = self._jvm.scala.Option.apply(hint)
        self._Check = self._Check.hasStandardDeviation(column, assertion_func, hint)
        return self

    def hasApproxCountDistinct(self, column, assertion, hint=None):
        """
        Creates a constraint that asserts on the approximate count distinct of the given column

        :param str column: Column in DataFrame to run the assertion on.
        :param lambda assertion: A function with an int or float parameter.
        :param str hint: A hint that states why a constraint could have failed.
        :return: hasApproxCountDistinct self: A Check object that asserts the  count distinct of the column.
        """
        assertion_func = ScalaFunction1(self._spark_session.sparkContext._gateway, assertion)
        hint = self._jvm.scala.Option.apply(hint)
        self._Check = self._Check.hasApproxCountDistinct(column, assertion_func, hint)
        return self

    def hasCorrelation(self, columnA, columnB, assertion, hint=None):
        """
        Creates a constraint that asserts on the pearson correlation between two columns.

        :param str columnA: First column in Data Frame which calculates the correlation.
        :param str columnB: Second column in Data Frame which calculates the correlation.
        :param lambda assertion: A function that accepts an int or float parameter.
        :param str hint: A hint that states why a constraint could have failed.
        :return: hasCorrelation self: A Check object that asserts the correlation calculation in the columns.
        """
        assertion_func = ScalaFunction1(self._spark_session.sparkContext._gateway, assertion)
        hint = self._jvm.scala.Option.apply(hint)
        self._Check = self._Check.hasCorrelation(columnA, columnB, assertion_func, hint)
        return self

    def satisfies(self, columnCondition, constraintName, assertion=None, hint=None):
        """
        Creates a constraint that runs the given condition on the data frame.

        :param str columnCondition: Data frame column which is a combination of expression and the column name.
                It has to comply with Spark SQL syntax. Can be written in an exact same way with conditions inside the
                `WHERE` clause.
        :param str constraintName: A name that summarizes the check being made. This name is being used to name the
                metrics for the analysis being done.
        :param lambda assertion: A function with an int or float parameter.
        :param str hint: A hint that states why a constraint could have failed.
        :return: satisfies self: A Check object that runs the condition on the data frame.
        """
        assertion_func = (
            ScalaFunction1(self._spark_session.sparkContext._gateway, assertion)
            if assertion
            else getattr(self._Check, "satisfies$default$3")()
        )
        hint = self._jvm.scala.Option.apply(hint)
        self._Check = self._Check.satisfies(
            columnCondition,
            constraintName,
            assertion_func,
            hint,
            self._jvm.scala.collection.Seq.empty(),
            self._jvm.scala.Option.apply(None)
        )
        return self

    def hasPattern(self, column, pattern, assertion=None, name=None, hint=None):
        """
        Checks for pattern compliance. Given a column name and a regular expression, defines a
        Check on the average compliance of the column's values to the regular expression.

        :param str column: Column in DataFrame to be checked
        :param Regex pattern: A name that summarizes the current check and the
                metrics for the analysis being done.
        :param lambda assertion: A function with an int or float parameter.
        :param str name: A name for the pattern constraint.
        :param str hint: A hint that states why a constraint could have failed.
        :return: hasPattern self: A Check object that runs the condition on the column.
        """
        if not assertion:
            assertion = is_one

        assertion_func = ScalaFunction1(self._spark_session.sparkContext._gateway, assertion)
        name = self._jvm.scala.Option.apply(name)
        hint = self._jvm.scala.Option.apply(hint)
        pattern_regex = self._jvm.scala.util.matching.Regex(pattern, None)
        self._Check = self._Check.hasPattern(
            column, pattern_regex, assertion_func, name, hint, self._jvm.scala.Option.apply(None)
        )
        return self

    def containsCreditCardNumber(self, column, assertion=None, hint=None):
        """
        Check to run against the compliance of a column against a Credit Card pattern.

        :param str column: Column in DataFrame to be checked. The column is expected to be a string type.
        :param lambda assertion: A function with an int or float parameter.
        :param hint hint: A hint that states why a constraint could have failed.
        :return: containsCreditCardNumber self: A Check object that runs the compliance on the column.
        """
        assertion = (
            ScalaFunction1(self._spark_session.sparkContext._gateway, assertion)
            if assertion
            else getattr(self._Check, "containsCreditCardNumber$default$2")()
        )
        hint = self._jvm.scala.Option.apply(hint)
        self._Check = self._Check.containsCreditCardNumber(column, assertion, hint)
        return self

    def containsEmail(self, column, assertion=None, hint=None):
        """
        Check to run against the compliance of a column against an e-mail pattern.

        :param str column: The Column in DataFrame to be checked. The column is expected to be a string datatype.
        :param lambda assertion: A function with an int or float parameter.
        :param str hint: A hint that states why a constraint could have failed.
        :return: containsCreditCardNumber self: A Check object that runs the compliance on the column.
        """
        assertion = (
            ScalaFunction1(self._spark_session.sparkContext._gateway, assertion)
            if assertion
            else getattr(self._Check, "containsEmail$default$2")()
        )
        hint = self._jvm.scala.Option.apply(hint)
        self._Check = self._Check.containsEmail(column, assertion, hint)
        return self

    def containsURL(self, column, assertion=None, hint=None):
        """
        Check to run against the compliance of a column against an e-mail pattern.

        :param str column: The Column in DataFrame to be checked. The column is expected to be a string datatype.
        :param lambda assertion: A function with an int or float parameter.
        :param str hint: A hint that states why a constraint could have failed.
        :return: containsURL self: A Check object that runs the compliance on the column.
        """
        assertion = (
            ScalaFunction1(self._spark_session.sparkContext._gateway, assertion)
            if assertion
            else getattr(self._Check, "containsURL$default$2")()
        )
        hint = self._jvm.scala.Option.apply(hint)
        self._Check = self._Check.containsURL(column, assertion, hint)
        return self

    def containsSocialSecurityNumber(self, column, assertion=None, hint=None):
        """
        Check to run against the compliance of a column against the Social security number pattern
        for the US.

        :param str column: The Column in DataFrame to be checked. The column is expected to be a string datatype.
        :param lambda assertion: A function with an int or float parameter.
        :param str hint: A hint that states why a constraint could have failed.
        :return: containsSocialSecurityNumber self: A Check object that runs the compliance on the column.
        """
        assertion = (
            ScalaFunction1(self._spark_session.sparkContext._gateway, assertion)
            if assertion
            else getattr(self._Check, "containsSocialSecurityNumber$default$2")()
        )
        hint = self._jvm.scala.Option.apply(hint)
        self._Check = self._Check.containsSocialSecurityNumber(column, assertion, hint)
        return self

    def hasDataType(self, column, datatype: ConstrainableDataTypes, assertion=None, hint=None):
        """
        Check to run against the fraction of rows that conform to the given data type.

        :param str column: The Column in DataFrame to be checked.
        :param ConstrainableDataTypes datatype: Data type that the columns should be compared against
        :param lambda assertion: A function with an int or float parameter.
        :param str hint: A hint that states why a constraint could have failed.

        :return: hasDataType self: A Check object that runs the compliance on the column.
        """
        datatype_jvm = datatype._get_java_object(self._jvm)
        assertion = (
            ScalaFunction1(self._spark_session.sparkContext._gateway, assertion)
            if assertion
            else getattr(self._Check, "hasDataType$default$3")()
        )
        hint = self._jvm.scala.Option.apply(hint)
        self._Check = self._Check.hasDataType(column, datatype_jvm, assertion, hint)
        return self

    def isNonNegative(self, column, assertion=None, hint=None):
        """
        Creates a constraint which asserts that a column contains no negative values.

        :param str column: The Column in DataFrame to run the assertion on.
        :param lambda assertion: A function with an int or float parameter.
        :param str hint: A hint that states why a constraint could have failed.

        :return: self (isNonNegative): A Check object that runs the compliance on the column.
        """
        assertion_func = (
            ScalaFunction1(self._spark_session.sparkContext._gateway, assertion)
            if assertion
            else getattr(self._Check, "isNonNegative$default$2")()
        )
        hint = self._jvm.scala.Option.apply(hint)
        self._Check = self._Check.isNonNegative(column, assertion_func, hint)
        return self

    def isPositive(self, column, assertion=None, hint=None):
        """
        Creates a constraint which asserts that a column contains no negative values and is greater than 0.

        :param str column: The Column in DataFrame to run the assertion on.
        :param lambda assertion: A function with an int or float parameter.
        :param str hint: A hint that states why a constraint could have failed.
        :return: isNonNegative self: A Check object that runs the assertion on the column.
        """
        assertion_func = (
            ScalaFunction1(self._spark_session.sparkContext._gateway, assertion)
            if assertion
            else getattr(self._Check, "isPositive$default$2")()
        )
        hint = self._jvm.scala.Option.apply(hint)
        self._Check = self._Check.isPositive(column, assertion_func, hint)
        return self

    def isLessThan(self, columnA, columnB, assertion=None, hint=None):
        """
        Asserts that, in each row, the value of columnA is less than the value of columnB

        :param str columnA: Column in DataFrame to run the assertion on.
        :param str columnB: Column in DataFrame to run the assertion on.
        :param lambda assertion: A function that accepts an int or float parameter.
        :param str hint: A hint that states why a constraint could have failed.
        :return: isLessThan self : A Check object that checks the assertion on the columns.
        """
        assertion_func = (
            ScalaFunction1(self._spark_session.sparkContext._gateway, assertion)
            if assertion
            else getattr(self._Check, "isLessThan$default$3")()
        )
        hint = self._jvm.scala.Option.apply(hint)
        self._Check = self._Check.isLessThan(columnA, columnB, assertion_func, hint)
        return self

    def isLessThanOrEqualTo(self, columnA, columnB, assertion=None, hint=None):
        """
        Asserts that, in each row, the value of columnA is less than or equal to the value of columnB.

        :param str columnA: Column in DataFrame to run the assertion on.
        :param str columnB: Column in DataFrame to run the assertion on.
        :param lambda assertion: A function that accepts an int or float parameter.
        :param str hint: A hint that states why a constraint could have failed.
        :return: isLessThanOrEqualTo self (isLessThanOrEqualTo): A Check object that checks the assertion on the columns.
        """
        assertion_func = (
            ScalaFunction1(self._spark_session.sparkContext._gateway, assertion)
            if assertion
            else getattr(self._Check, "isLessThanOrEqualTo$default$3")()
        )
        hint = self._jvm.scala.Option.apply(hint)
        self._Check = self._Check.isLessThanOrEqualTo(columnA, columnB, assertion_func, hint)
        return self

    def isGreaterThan(self, columnA, columnB, assertion=None, hint=None):
        """
        Asserts that, in each row, the value of columnA is greater than the value of columnB

        :param str columnA: Column in DataFrame to run the assertion on.
        :param str columnB: Column in DataFrame to run the assertion on.
        :param lambda assertion: A function that accepts an int or float parameter.
        :param str hint: A hint that states why a constraint could have failed.
        :return: isGreaterThan self: A Check object that runs the assertion on the columns.
        """
        assertion_func = (
            ScalaFunction1(self._spark_session.sparkContext._gateway, assertion)
            if assertion
            else getattr(self._Check, "isGreaterThan$default$3")()
        )
        hint = self._jvm.scala.Option.apply(hint)
        self._Check = self._Check.isGreaterThan(columnA, columnB, assertion_func, hint)
        return self

    def isGreaterThanOrEqualTo(self, columnA, columnB, assertion=None, hint=None):
        """
        Asserts that, in each row, the value of columnA is greather than or equal to the value of columnB

        :param str columnA: Column in DataFrame to run the assertion on.
        :param str columnB: Column in DataFrame to run the assertion on.
        :param lambda assertion: A function that accepts an int or float parameter.
        :param str hint: A hint that states why a constraint could have failed.
        :return: isGreaterThanOrEqualTo self: A Check object that runs the assertion on the columns.
        """
        assertion_func = (
            ScalaFunction1(self._spark_session.sparkContext._gateway, assertion)
            if assertion
            else getattr(self._Check, "isGreaterThanOrEqualTo$default$3")()
        )
        hint = self._jvm.scala.Option.apply(hint)
        self._Check = self._Check.isGreaterThanOrEqualTo(columnA, columnB, assertion_func, hint)
        return self

    def isContainedIn(self, column, allowed_values, assertion=None, hint=None):
        """
        Asserts that every non-null value in a column is contained in a set of predefined values

        :param str column: Column in DataFrame to run the assertion on.
        :param list[str] allowed_values: A function that accepts allowed values for the column.
        :param lambda assertion: A function that accepts an int or float parameter.
        :param str hint: A hint that states why a constraint could have failed.
        :return: isContainedIn self: A Check object that runs the assertion on the columns.
        """
        arr = self._spark_session.sparkContext._gateway.new_array(self._jvm.java.lang.String, len(allowed_values))
        for i in range(len(allowed_values)):
            arr[i] = allowed_values[i]

        if not assertion:
            assertion = is_one
        assertion_func = ScalaFunction1(self._spark_session.sparkContext._gateway, assertion)
        hint = self._jvm.scala.Option.apply(hint)
        self._Check = self._Check.isContainedIn(column, arr, assertion_func, hint)
        return self

    def evaluate(self, context):
        """
        Evaluate this check on computed metrics.

        :param context: result of the metrics computation
        :return: evaluate self: A Check object that evaluates the check.
        """
        raise NotImplementedError("We need an AnalyzerContext wrapper class before implementing")

    def requiredAnalyzers(self):
        # TODO: documentation
        raise NotImplementedError("Need to evaluate usage first, seems to be called internally by Checks")



# -*- coding: utf-8 -*-
"""
Repository file for all the different metrics repository classes in Deequ

Author: Calvin Wang
"""
import json
import time

from pyspark.sql import DataFrame, SparkSession

from pydeequ.scala_utils import to_scala_map, to_scala_seq


class MetricsRepository:
    """
    Base class for Metrics Repository
    """

    @classmethod
    def helper_metrics_file(cls, spark_session: SparkSession, filename: str = "metrics.json"):
        """
        Helper method to create the metrics file for storage
        """
        tempDir = spark_session._jvm.com.google.common.io.Files.createTempDir()
        f = spark_session._jvm.java.io.File(tempDir, filename)
        return f.getAbsolutePath()

    def _check_RepositoryLoader(self):
        """
        Private method for checking if our repository loader has been initialized yet.
        Repository loader is the object that will handle the loaded metrics repository runs
        from the past and prompt for further filtering of the below methods.
        """
        if not self.RepositoryLoader:
            raise AttributeError("No data loaded yet, did you run repository.load() yet?")

    def load(self):
        """
        Get a builder class to construct a loading query to get AnalysisResults
        """
        self.RepositoryLoader = self.repository.load()
        return self

    def withTagValues(self, tagValues: dict):
        """
        Filter out results that don't have specific values for specific tags
        :param tagValues: Dict with tag names and the corresponding values to filter for
        """
        self._check_RepositoryLoader()
        self.RepositoryLoader.withTagValues(to_scala_map(self._spark_session, tagValues))
        return self

    def forAnalyzers(self, analyzers: list):
        """
        Choose all metrics that you want to load
        :param analyzers: List of analyers who's resulting metrics you want to load
        """
        analyzers_jvm = []
        for analyzer in analyzers:
            analyzer._set_jvm(self._jvm)
            analyzers_jvm.append(analyzer._analyzer_jvm)

        self.RepositoryLoader.forAnalyzers(to_scala_seq(self._jvm, analyzers_jvm))
        return self

    def before(self, dateTime: int):
        """
        Only look at AnalysisResults with a result key with a smaller value
        :param dateTime: The maximum dateTime of AnalysisResults to look at
        """
        self._check_RepositoryLoader()
        self.RepositoryLoader.before(dateTime)
        return self

    def after(self, dateTime: int):
        """
        Only look at AnalysisResults with a result key with a greater value
        :param dateTime: The minimum dateTime of AnalysisResults to look at
        """
        self._check_RepositoryLoader()
        self.RepositoryLoader.after(dateTime)
        return self

    def getSuccessMetricsAsJson(self, withTags: list = None):
        """
        Get the AnalysisResult as JSON
        :param withTags: List of tags to filter previous Metrics Repository runs with
        """
        self._check_RepositoryLoader()
        if not withTags:
            withTags = getattr(self.repository.load(), "getSuccessMetricsAsJson$default$1")()  # empty sequence
        else:
            withTags = to_scala_seq(self._jvm, withTags)
        return json.loads(self.RepositoryLoader.getSuccessMetricsAsJson(withTags))

    def getSuccessMetricsAsDataFrame(self, withTags: list = None, pandas: bool = False):
        """
        Get the AnalysisResult as DataFrame
        :param withTags: List of tags to filter previous Metrics Repository runs with
        """
        self._check_RepositoryLoader()
        if not withTags:
            withTags = getattr(self.repository.load(), "getSuccessMetricsAsDataFrame$default$2")()  # empty sequence
        else:
            withTags = to_scala_seq(self._jvm, withTags)
        success = self.RepositoryLoader.getSuccessMetricsAsDataFrame(self._jspark_session, withTags)
        return DataFrame(success, self._spark_session).toPandas() if pandas else DataFrame(success, self._spark_session)


class InMemoryMetricsRepository(MetricsRepository):
    """
    High level InMemoryMetricsRepository Interface
    """

    def __init__(self, spark_session: SparkSession):
        self._spark_session = spark_session
        self._jvm = spark_session._jvm
        self._jspark_session = spark_session._jsparkSession
        self.deequIMmetRep = spark_session._jvm.com.amazon.deequ.repository.memory.InMemoryMetricsRepository
        self.repository = self.deequIMmetRep()


class FileSystemMetricsRepository(MetricsRepository):
    """
    High level FileSystemMetricsRepository Interface

    :param spark_session: SparkSession
    :param path: The location of the file metrics repository.
    """

    def __init__(self, spark_session: SparkSession, path: str = None):

        self._spark_session = spark_session
        self._jvm = spark_session._jvm
        self._jspark_session = spark_session._jsparkSession
        if not path:
            path = self.helper_metrics_file(self._spark_session)
        self.path = path
        self.deequFSmetRep = spark_session._jvm.com.amazon.deequ.repository.fs.FileSystemMetricsRepository
        self.repository = self.deequFSmetRep(self._jspark_session, path)


class ResultKey:
    """
    Information that uniquely identifies a AnalysisResult

    :param spark_session: SparkSession
    :param dataSetDate: Date of the result key
    :param tags: A map with additional annotations

    """

    def __init__(self, spark_session: SparkSession, dataSetDate: int = None, tags: dict = None):

        self.resultKey = self.__key(spark_session, dataSetDate, tags)

    def __key(self, spark_session: SparkSession, dataSetDate: int = None, tags: dict = None):
        """
        Private method for key construction.

        :param spark_session: Spark Session
        :param dataSetDate: A date related to the AnalysisResult
        :param tags: A map with additional annotations
        """
        if not dataSetDate:
            dataSetDate = self.current_milli_time()
        if not tags:
            tags = getattr(
                spark_session._jvm.com.amazon.deequ.repository.ResultKey, "apply$default$2"
            )()  # empty scala map
        else:
            tags = to_scala_map(spark_session, tags)

        return spark_session._jvm.com.amazon.deequ.repository.ResultKey(dataSetDate, tags)

    @staticmethod
    def current_milli_time():
        """
        Get current time in milliseconds
        # TODO: Consider putting this into scala_utils? Or general utils?
        """
        return int(round(time.time() * 1000))
